import unittest 
import sys
import os

from cromulent import vocab, model
from cromulent.model import factory

class TestClassBuilder(unittest.TestCase):
	def setUp(self):
		pass

	def tearDown(self):
		pass

	def test_class(self):
		vocab.register_aat_class("TestObject1", {"parent": model.HumanMadeObject, "id": "1", "label": "example 1"})
		from cromulent.vocab import TestObject1
		self.assertEqual(TestObject1._classification[0].id, 'http://vocab.getty.edu/aat/1')

	def test_instance(self):
		vocab.register_instance("TestMaterial2", {"parent": model.Material, "id": "2", "label": "example 2"})
		self.assertTrue('TestMaterial2' in vocab.instances)
		tm2 = vocab.instances['TestMaterial2']
		self.assertEqual(tm2.id, "http://vocab.getty.edu/aat/2")

	def test_metatype(self):
		vocab.register_instance("example", {"parent": model.Type, "id": "3", "label": "example type"}) 
		vocab.register_aat_class("TestObject2", 
			{"parent": model.HumanMadeObject, "id": "4", "label": "example typed object", "metatype": "example"})
		from cromulent.vocab import TestObject2
		self.assertEqual(TestObject2._classification[0].classified_as[0].id, 'http://vocab.getty.edu/aat/3')

	def test_multitype(self):
		from cromulent.vocab import make_multitype_obj, Painting, Drawing
		inst = make_multitype_obj(Painting, Drawing)
		self.assertTrue(isinstance(inst, Painting))
		self.assertTrue(len(inst.classified_as) == 2)
		self.assertTrue(inst.classified_as[1].id == "http://vocab.getty.edu/aat/300033973")

		from cromulent.model import HumanMadeObject

		inst = make_multitype_obj(HumanMadeObject, Painting)
		self.assertTrue(len(inst.classified_as) == 1)
		self.assertTrue(inst.classified_as[0].id == "http://vocab.getty.edu/aat/300033618")

	def test_conceptual_parts(self):
		r = model.Right()
		r2 = model.Right()
		self.assertRaises(model.DataError, r.__setattr__, 'part', r2)
		r.conceptual_part = r2
		self.assertTrue(r2 in r.conceptual_part)

		vocab.conceptual_only_parts()
		r3 = model.Right()
		r4 = model.Right()
		r3.part = r4
		self.assertTrue(r4 in r3.conceptual_part)
		self.assertTrue("part" in model.factory.toJSON(r3))
		self.assertTrue(r4 in r3.part)


	def test_art_setter(self):
		p = model.HumanMadeObject("a", art=1)
		p._label = "a"
		pj = p._toJSON(done={})
		self.assertFalse(pj.get('classified_as', None))
		vocab.add_art_setter()
		p2 = vocab.Painting("b", art=1)
		p2j = p2._toJSON(done={})

	def test_aa_check(self):

		# Make sure that some other test hasn't set it
		try:
			del model.AttributeAssignment.set_assigned_property
		except:
			pass

		t = model.Type()
		aa = model.AttributeAssignment()
		# First check that aa accepts a type
		aa.assigned_property = t
		# And will not accept a string
		self.assertRaises(model.DataError, aa.__setattr__, "assigned_property", "classified_as")

		# Check we can set anything to assigned / assigned_to
		aa.assigned_property = None
		aa.assigned = aa
		aa.assigned_to = aa
		self.assertEqual(aa.assigned[0], aa)
		self.assertEqual(aa.assigned_to, aa)

		vocab.add_attribute_assignment_check()

		# This should fail right now as can't classify as an AA
		self.assertRaises(model.DataError, aa.__setattr__, "assigned_property", "classified_as")
		aa.assigned = None
		aa.assigned_to = None
		aa.assigned = t
		aa.assigned_to = t
		aa.assigned_property = "classified_as"
		self.assertEqual(aa.assigned_property, 'classified_as')


	def test_boundary_setter(self):
		vocab.add_linked_art_boundary_check()
		p = model.Person()
		p2 = model.Person()
		n = model.Name()
		n.content = "Test"
		p2.identified_by = n
		p.equivalent = p2
		# Now, Test should not appear in the resulting JSON of p
		factory.linked_art_boundaries = True
		js = factory.toJSON(p)
		self.assertTrue(not 'identified_by' in js['equivalent'][0])
		factory.linked_art_boundaries = False
		js = factory.toJSON(p)
		self.assertTrue('identified_by' in js['equivalent'][0])		

	def test_procurement_boundary(self):
		vocab.add_linked_art_boundary_check()
		a = model.Activity()
		p = vocab.ProvenanceEntry()
		a.caused = p
		js = factory.toJSON(a)
		self.assertTrue(not 'classified_as' in js['caused'][0])		

	def test_linguistic_object_boundary(self):
		vocab.add_linked_art_boundary_check()
		jrnl = vocab.JournalText(label="journal")
		issue = vocab.IssueText(label="issue")
		issue.part_of = jrnl
		issue.referred_to_by = vocab.MaterialStatement(content="Statement")

		js = factory.toJSON(issue)
		# Have not embedded journal in issue
		self.assertTrue(not 'classified_as' in js['part_of'][0])
		# Have embedded statement in issue
		self.assertTrue('content' in js['referred_to_by'][0])
		self.assertTrue('type' in js['referred_to_by'][0]['classified_as'][0]['classified_as'][0])





import unittest 

try:
	from collections import OrderedDict
except:
	# 2.6
	from ordereddict import OrderedDict

from cromulent import reader
from cromulent.model import factory, Person, DataError, BaseResource, \
	Dimension, override_okay, AttributeAssignment

from cromulent import vocab

class TestReader(unittest.TestCase):

	def setUp(self):
		self.reader = reader.Reader()
		# ensure we can use parent_of
		override_okay(Person, 'parent_of')
		# Person._properties['parent_of']['multiple'] = 1

	def test_read(self):
		self.assertRaises(DataError, self.reader.read, "")
		self.assertRaises(DataError, self.reader.read, "This is not JSON")
		self.assertRaises(DataError, self.reader.read, "{}")

		whostr = '{"type": "Person", "_label": "me"}'
		self.assertTrue(isinstance(self.reader.read(whostr), Person))

		whostr = '{"@context": "fishbat", "type": "Person", "_label": "me"}'
		self.assertTrue(isinstance(self.reader.read(whostr), Person))

		levelstr = '{"type": "Person", "parent_of": {"type": "Person", "_label": "child"}}'
		self.assertTrue(isinstance(self.reader.read(levelstr).parent_of[0], Person))

		basestr = '{"_label": "base"}'
		self.assertTrue(isinstance(self.reader.read(basestr), BaseResource))

		unknown = '{"type":"FishBat"}'
		self.assertRaises(DataError, self.reader.read, unknown)

		unknown2 = '{"type":"Person", "fishbat": "bob"}'
		self.assertRaises(DataError, self.reader.read, unknown)

	def test_attrib_assign(self):
		vocab.add_attribute_assignment_check()

		data = """
		{
		  "id": "https://linked.art/example/activity/12", 
		  "type": "AttributeAssignment", 
		  "assigned": {
		    "id": "https://linked.art/example/name/10", 
			"type": "Name", 
		    "content": "Exhibition Specific Name"
		  }, 
		  "assigned_property": "identified_by", 
		  "assigned_to": {
		    "id": "https://linked.art/example/object/12", 
		    "type": "HumanMadeObject", 
		    "_label": "Real Painting Name"
		  }
		}
		"""
		d = self.reader.read(data)
		self.assertTrue(isinstance(d, AttributeAssignment))


	def test_vocab_collision(self):
		# Test that the algorithm picks the right vocab instance
		# if multiple have the same AAT term but different base class

		data = """
        {
          "type": "LinguisticObject",
          "_label": "Sale recorded in catalog: B-267 0003 (1817) (record number 22947)",
	      "part_of": [
            {
              "type": "LinguisticObject",
              "_label": "Sale Catalog B-267",
              "classified_as": [
                {
                  "id": "http://vocab.getty.edu/aat/300026068",
                  "type": "Type",
                  "_label": "Auction Catalog"
                }
              ]
            }
          ]
        }
        """
		d = self.reader.read(data)
		self.assertTrue(isinstance(d.part_of[0], vocab.AuctionCatalogText))





from __future__ import unicode_literals

import os
import sys
import re
import codecs
import inspect
import uuid
import datetime
import json
from json import JSONEncoder
from collections import OrderedDict
from collections import namedtuple
from pyld import jsonld
try:
	from rdflib import Dataset as CG
except:
	from rdflib import ConjunctiveGraph as CG

KEY_ORDER_DEFAULT = 10000
LINKED_ART_CONTEXT_URI = "https://linked.art/ns/v1/linked-art.json"

try:
	STR_TYPES = [str, unicode] #Py2.7
	FILE_STREAM_CLASS = file
except:
	import io
	STR_TYPES = [bytes, str] #Py3.x
	FILE_STREAM_CLASS = io.TextIOBase


pyld_proc = jsonld.JsonLdProcessor()
min_context = {
	"crm": "http://www.cidoc-crm.org/cidoc-crm/",
    "sci": "http://www.ics.forth.gr/isl/CRMsci/",
    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
    "dc": "http://purl.org/dc/elements/1.1/",
    "dcterms": "http://purl.org/dc/terms/",
    "schema": "http://schema.org/",
    "skos": "http://www.w3.org/2004/02/skos/core#",
    "foaf": "http://xmlns.com/foaf/0.1/",
    "xsd": "http://www.w3.org/2001/XMLSchema#",
    "dig": "http://www.ics.forth.gr/isl/CRMdig/",
    "la": "https://linked.art/ns/terms/",
    "id": "@id",
    "type": "@type"
}
re_bnodes = re.compile("^_:b([0-9]+) ", re.M)
re_bnodeo = re.compile("> _:b([0-9]+) <", re.M)
re_quad = re.compile(" <[^<]+?> .$", re.M)

PropInfo = namedtuple("PropInfo", [
	'property', # the name of the property, eg 'identified_by'
	'predicate', # the name from the ontology, eg 'crm:P1_is_identified_by'
	'range', # the class which is the range of this predicate, eg model.Identifier
	'inverse_property', # the name of the property which is the inverse of this one, eg 'identifies'
	'inverse_predicate', # the predicate for the inverse of this one, eg 'crm:P1i_idenfifies'
	'multiple_okay', # can there be multiple values, eg True
	'profile_okay' # is this property okay according to the profile, eg 0 (no), 1 (yes), 2 (warn)
	])


class CromulentError(Exception):
	"""Base exception class"""

	resource = None

	def __init__(self, msg, resource=None):
		"""Initialize CidocError."""
		self.args = [msg]
		self.resource = resource

class ConfigurationError(CromulentError):
	"""Raised when an object (likely the factory) isn't configured properly for the current operation."""
	pass

class MetadataError(CromulentError):
	"""Base metadata exception."""
	pass

class RequirementError(MetadataError):
	"""Raised when schema/profile/metadata requirements not met."""
	pass

class DataError(MetadataError):
	"""Raised when data is not valid/allowed."""
	pass

class ProfileError(MetadataError):
	"""Raised when a class or property not in the configured profile is used"""
	pass

class CromJsonEncoder(JSONEncoder):

	def default(self, o):
		if isinstance(o, BaseResource):
			# print("Saw %r" % o)
			return o._minToJSON()
		else:
			return JSONEncoder.default(self, o)			

class CromulentFactory(object):

	def __init__(self, base_url="", base_dir="", lang="", full_names=False, 
		context="", context_file={}, load_context=True):
		self.base_url = base_url
		self.base_dir = base_dir

		self.debug_level = "warn"
		self.log_stream = sys.stderr

		self.materialize_inverses = False # Create the inverse relationship on the object to the subject
		self.full_names = False # Use full property and class names in output
		self.pipe_scoped_contexts = False # Serialize to "name|Pnn_x_y" for scoped properties (documentation)
		self.validate_properties = True # validate properties are even CRM
		self.validate_profile = True # Validate linked_art specific stuff
		self.validate_range = True # Raise if attempt to set prop to invalid class
		self.validate_multiplicity = True  # Raise if attempt to set n:1 to []
		self.auto_assign_id = True # Automatially assign a URI
		self.process_multiplicity = True # Return multiple with single value as [value]
		self.multiple_instances_per_property = "drop"
		self.allow_highlight = False # Allow the JSON to include a _highlight flag for re-rendering
		self.allow_elide = False

		self.pair_tree_levels = 0

		self.auto_id_type = "int-per-segment" #  "int", "int-per-type", "int-per-segment", "uuid", "uuid-segment"
		# self.default_lang = lang  # NOT USED
		self.filename_extension = ".json"  # some people like .jsonld
		self.context_uri = context # Might be a list, or a context value as a dict
		self.context_json = {}

		self.prefixes = {}
		self.prefixes_rev = {}
		self.context_rev = {}
		# Maybe load it up for prefixes
		if load_context:
			# Leave this as a map for future extensions
			context_filemap = {
				LINKED_ART_CONTEXT_URI: 
					os.path.join(os.path.dirname(__file__), 'data', 'linked-art.json')
			}
			context_filemap.update(context_file)
			self.load_context(context, context_filemap)

		self.elasticsearch_compatible = False # return {'id': 'uri'} instead of string
		self.linked_art_boundaries = False # break on linked art API boundaries between classes
		self.id_type_label = True # references are id, type and _label, not just id.

		# if sorting is unimportant, use fast. If sorting is important, and python >= 3.6, use fast.
		# fast is approximately half the time for serializing
		self.json_serializer = "normal" # "normal" or "fast"
		self.json_indent = 2
		self.order_json = True
		self.key_order_hash = {"@context": 0, "id": 1, "type": 2, 
			"_label": 5, "value": 6}
		self.full_key_order_hash = {"@context": 0, "@id": 1, "rdf:type": 2, "@type": 2,
			"rdfs:label": 5, "rdf:value": 6}
		self.key_order_default = 10000

		self.underscore_properties = ["_label"]

		self._auto_id_types = {}
		self._auto_id_segments = {}
		self._auto_id_int = -1
		self._all_classes = {}

	def load_context(self, context, context_filemap):
		if not context or not context_filemap:
			raise ConfigurationError("No context provided, and load_context not False")

		if type(context) is not list:
			context = [context]

		js = {'@context': {}}
		for ct in context:
			fn = context_filemap.get(ct, "")
			if fn:
				try:
					fh = open(fn)
					data = fh.read()
					fh.close()
				except IOError:
					raise ConfigurationError("Provided context file does not exist")
			else:
				# XXX Fetch from web
				data = "{}"

			try:
				ctx = json.loads(data)
				js['@context'].update(ctx['@context'])
			except:
				raise ConfigurationError("Provided context does not have valid JSON")				
		# this is the merged context information, not any single one
		self.context_json = js
		self.process_context()

	def process_context(self):
		# Filter context looking for prefixes
		# And make reverse mapping
		# Note that this does not process scoped contexts for member_of / part_of
		for (k,v) in self.context_json['@context'].items():
			if type(v) in STR_TYPES and v[-1] in ['/', '#']:
				self.prefixes[k] = v
				self.prefixes_rev[v] = k
			elif k == "@version":
				continue
			else:
				if type(v) in STR_TYPES:
					rdf = v
				else:
					rdf = v['@id']
				self.context_rev[rdf] = k

	def __getstate__(self):
		# Make a copy of current object state
		d = self.__dict__.copy()
		# try to flush the stream
		try:
			self.log_stream.flush()
		except:
			# stream instance may not support flush()
			pass
		# Now reify the log stream
		strm = d['log_stream']
		if strm is sys.stdout:
			d['log_stream'] = ("sys.stdout", "stream")
		elif strm is sys.stderr:
			d['log_stream'] = ("sys.stderr", "stream")
		elif isinstance(strm, FILE_STREAM_CLASS):
			d['log_stream'] = (strm.name, "file")
		else:
			d['log_stream'] = None
		return d

	def __setstate__(self, state):
		# State is __dict__ with a reified log_stream as above
		self.__dict__.update(state)
		if self.log_stream:
			if self.log_stream[1] == "stream":
				if self.log_stream[0] == "sys.stdout":
					self.log_stream = sys.stdout
				elif self.log_stream[0] == "sys.stderr":
					self.log_stream = sys.stderr
			elif self.log_stream[1] == "file":
				try:
					self.log_stream = open(self.log_stream[0], 'a') 
				except:
					self.log_stream = None


	def set_debug_stream(self, strm):
		"""Set debug level."""
		self.log_stream = strm

	def set_debug(self, typ):
		"""Set behavior on errors and warnings.

		error = squash warnings
		warn = display warnings
		error_on_warning = raise exception for a warning rather than continuing
		"""
		if typ in ['error', 'warn', 'error_on_warning']:
			self.debug_level = typ
		else:
			raise ConfigurationError("Only levels are 'error', 'warn' and 'error_on_warning'")

	def maybe_warn(self, msg):
		"""warn method that respects debug_level property."""
		if self.log_stream and self.debug_level == "warn":
			self.log_stream.write(msg + "\n")
			try:	
				self.log_stream.flush()
			except:
				pass
		elif self.debug_level == "error_on_warning":
			# We don't know the type, just raise a MetadataError
			raise MetadataError(msg)

	def _is_uri(self, what):
		uri_schemes = ['urn:uuid:', 'tag:', 'data:', 'mailto:', 'info:', 'ftp:/', 'sftp:/'] 
		for u in uri_schemes:
			if what.startswith(u):
				return True
		return False

	def generate_id(self, what, auto_type=None):
		if auto_type == None:
			auto_type = self.auto_id_type

		if auto_type == "int":
			# increment and return
			self._auto_id_int += 1
			slug = self._auto_id_int
		elif auto_type == "int-per-segment":
			curr = self._auto_id_segments.get(what._uri_segment, -1)
			curr += 1
			self._auto_id_segments[what._uri_segment] = curr
			slug = self._auto_id_segments[what._uri_segment]
		elif auto_type == "int-per-type":
			t = type(what).__name__
			curr = self._auto_id_types.get(t, -1)
			curr += 1
			self._auto_id_types[t] = curr
			slug = self._auto_id_types[t]
		elif auto_type == "uuid":
			return "urn:uuid:%s" % uuid.uuid4()
		elif auto_type == "uuid-segment":
			slug = uuid.uuid4()
		else:
			raise ConfigurationError("Unknown auto-id type")

		if what.__class__._uri_segment:
			seg = what.__class__._uri_segment + "/"
		else:
			seg = ""

		return self.base_url + seg + str(slug)	

	def find_serializable(self, what):

		if not self.linked_art_boundaries:
			raise ConfigurationError("Factory doesn't have any boundaries to distinguish between entities")

		found = []
		props = what.list_my_props()
		for p in props:
			if p in ['id', 'type', '_label', 'content', 'value', 'begin_of_the_begin', 'end_of_the_end']:
				continue
			val = getattr(what, p)
			if isinstance(val, ExternalResource):
				val = [val]
			if type(val) is list:
				for v in val:
					if isinstance(v, ExternalResource):
						if not v in found and v.id and not v._linked_art_boundary_okay(what, p, v) and set(v.list_my_props()).difference(set(["_label", "id"])):
							found.append(v)
						downstream = self.find_serializable(v)
						for d in downstream:
							if not d in found:
								found.append(d)
		return found

	def toJSON(self, what, done=None):
		""" Serialize what, making sure of no infinite loops """
		if not done:
			done = {}
		if self.json_serializer == "fast":
			out = what._toJSON_fast(top=what, done=done)
		else:
			out = what._toJSON(top=what, done=done)
		return out

	def _collapse_json(self, text, collapse):
		js_indent = self.json_indent
		lines = text.splitlines()
		out = [lines[0]]
		while lines:
			l = lines.pop(0)
			indent = len(re.split('[^ \t\n\r]', l, 1)[0])
			if indent and l.rstrip()[-1] in ['[', '{']:
				curr = indent
				temp = []
				stemp = []
				while lines and curr <= indent:
					if temp and curr == indent:
						break
					temp.append(l[curr:])
					stemp.append(l.strip())
					l = lines.pop(0)
					indent = len(re.split('[^ \t\n\r]', l, 1)[0])
				temp.append(l[curr:])
				stemp.append(l.lstrip())

				short = " " * curr + ''.join(stemp)
				if len(short) < collapse:
					out.append(short)
				else:
					ntext = '\n'.join(temp)
					nout = self._collapse_json(ntext, collapse)					
					for no in nout:
						out.append(" " * curr + no)
			elif indent:
				out.append(l)
		out.append(l)
		return out
 
	def collapse_json(self, text, collapse):
		return '\n'.join(self._collapse_json(text, collapse))

	def _buildString(self, js, compact=True, collapse=0):
		"""Build string from JSON."""
		try:
			if compact:
				out = json.dumps(js, separators=(',',':'), ensure_ascii=False)
			else:
				out = json.dumps(js, indent=self.json_indent, ensure_ascii=False)
		except:
			out = ""
			self.maybe_warn("Can't decode %r" % js)
			raise
		if collapse:
			out = self.collapse_json(out, collapse)
		return out 		

	def toString(self, what, compact=True, collapse=0, done=None):
		"""Return JSON setialization as string."""
		if not done:
			done = {}
		js = self.toJSON(what, done=done)
		return self._buildString(js, compact, collapse)


	def toHtml(self, what, done=None):
		enc = JSONEncoder(indent=self.json_indent, ensure_ascii=False)
		js = self.toJSON(what, done=done)
		res = ['<pre><span>']
		lineno = 0
		for y in enc.iterencode(js):
			# split newlines
			if '\n' in y:
				sp = y.split('\n')
				x2 = []
				for z in sp[:-1]:
					if z:
						x2.append(z)
					x2.append('\n')
				if sp[-1]:
					x2.append(sp[-1])
			else:
				x2 = [y]

			for x in x2:
				if x[0] == '"':
					s = x.strip()[1:-1]
					clss = "str"
					if s.startswith('http') or self._is_uri(s):
						x = '"<a href="%s">%s</a>"' % (s, s)
					elif s in ["@context"]:
						clss = "str context"
					res.append('<span class="%s">%s</span>' % (clss, x))
				elif x[0].isdigit():
					res.append('<span class="int">%s</span>' % x)						
				elif x[0] in [":", ","]:
					res.append('<span class="sep">%s</span>' % x)	
				elif x[0]  == "{":
					res.append('<span class="bsep">%s</span><span class="block">' % x)
				elif x[0]  == "[":
					res.append('<span class="asep">%s</span><span class="ablock">' % x)
				elif x[0] == "}":
					res.append('</span><span class="bsep">%s</span>' % x)
				elif x[0] == "]":
					res.append('</span><span class="asep">%s</span>' % x)
				elif x == '\n':
					res.append('\n<span class="line" data-lno="%s"></span>' % lineno)
					lineno += 1
				elif x.isspace():
					res.append(x)
				else:
					print("*** Unhandled: %r" % x)
					res.append(x)
		res.append('</span></pre>')
		return ''.join(res)

	def toRDF(self, what, format="nq", bnode_prefix=""):
		# Format can be:  xml, pretty-xml, turtle, n3, nt, trix, trig, nquads
		# ttl = turtle; nq, n-quads == nquads

		# Need to ensure we generate the full form of predicates
		# otherwise context processing takes AGES
		# So set serializer to normal, and full_names to True
		srlz = self.json_serializer
		fn = self.full_names
		self.json_serializer = "normal"
		self.full_names = True
		js  = self.toJSON(what)
		# And put them back
		self.json_serializer = srlz
		self.full_names = fn

		# Substitute in a minimal context that defines only prefixes
		js['@context'] = min_context
		src = {'@id': js['@id'], '@graph':js}
		data = pyld_proc.to_rdf(src, options={"format": "application/nquads"})

		# Here replace all the bnodes with a unique id
		# This works so long as PyLD continues with incrementing integer bnode ids
		if bnode_prefix:
			data = re_bnodes.subn("_:b{bnode_prefix}_\\1 ".format(bnode_prefix=bnode_prefix), data)[0]
			data = re_bnodeo.subn("> _:b{bnode_prefix}_\\1 <".format(bnode_prefix=bnode_prefix), data)[0]				

		if format in ['nq', 'nquads', 'n-quads', 'application/nquads']:
			return data
		elif format in ['nt', 'ntriples', 'n-triples', 'application/ntriples']:
			data = re_quad.subn(" .", data)[0]
			return data
		else:
			# Need to pass over to rdflib
			g = CG()
			for (k,v) in min_context.items():
				if v[0] != "@":
					g.bind(k, v)
			g.parse(data=data, format="nquads")
			out = g.serialize(format=format)
			if type(out) == bytes:
				return out.decode('utf-8')
			else:
				return out

	def get_filename(self, whatid, extension=""):

		mdb = self.base_url
		if not whatid.startswith(mdb):
			raise ConfigurationError("The id of that object is not the base URI (factory.base_url) in the Factory")
		mdd = self.base_dir
		if not mdd:
			raise ConfigurationError("Directory (factory.base_dir) on Factory must be set to generate a file name")
		fp = whatid[len(mdb):]	

		# This will always be /, as it's from the URI
		bits = fp.split('/')
		fn = bits[-1]
		dirs = bits[:-1]

		if self.pair_tree_levels:
			for d in range(self.pair_tree_levels):
				if len(fn) > 2*d+1:
					dirs.append(fn[2*d:2*d+2])

		if len(dirs):
			mydir = os.path.join(mdd, *dirs)		
			try:
				os.makedirs(mydir)
			except OSError:
				pass

		# Allow passing in an override
		if extension:
			fn = fn + extension
		elif self.filename_extension:
			fn = fn + self.filename_extension
		dirs.append(fn)
		filename = os.path.join(mdd, *dirs)		
		return filename

	def toFile(self, what, compact=True, filename="", done=None, format=None, bnode_prefix="", extension=""):
		"""Write to local file.

		Creates directories as necessary based on URI, if filename is not supplied
		"""

		if not done:
			done = {}

		if not format:
			if not filename:
				filename = self.get_filename(what.id, extension=extension)
			js = self.toJSON(what, done=done)
			out = self._buildString(js, compact)
		else:
			if not filename:
				if extension:
					ext = extension
				elif format == "pretty-xml":
					ext = "xml"
				else:
					ext = format
				filename = self.get_filename(what.id, extension=ext)
			out = self.toRDF(what, format=format, bnode_prefix=bnode_prefix)

		fh = open(filename, 'w')
		try:
			fh.write(out)
		except:
			# Could be 2.x unicode issue
			fh.write(out.encode('utf-8'))
		fh.close()
		return out

	def production_mode(self, state=True):
		if state:
			self.cache_hierarchy()
			self.validate_profile = False
			self.validate_properties = False
			self.validate_range = False
			self.validate_multiplicity = False
			return True
		else:
			self.validate_profile = True
			self.validate_properties = True
			self.validate_range = True
			self.validate_multiplicity = True
			# no way to undo the hierarchy cache
			return False
			
	def cache_hierarchy(self):
		""" For each class, walk up the hierarchy and cache the terms """
		# This will work with the existing code, as it will find it in the first
		# test of props on classhier[0], the loop will just terminate straight away

		for c in self._all_classes.values():
			new_hash = c._all_properties.copy()
			if len(c._classhier) > 1:
				for p in c._classhier[1:]:
					for (prop, info) in p._all_properties.items():
						if not prop in new_hash:
							new_hash[prop] = info
			c._all_properties = new_hash


class ExternalResource(object):
	"""Base class for all resources, including external references"""
	
	_factory = None
	_uri_segment = ""
	id = ""
	_full_id = ""
	_all_properties = {}
	_type = ""
	_embed = True
	_property_name_map = {}
	_highlight = False
	_elide = False


	def __init__(self, ident=None):
		self._factory = factory
		if ident is not None:
			if self._factory._is_uri(ident):
				self.id = ident
			elif ident.startswith('http'):
				# Try to find prefixable term
				hashed = ident.rsplit('#', 1)
				if len(hashed) == 1:
					(pref, rest) = ident.rsplit('/', 1)
					pref += "/"
				else:
					(pref, rest) = hashed
					pref += "#"

				if pref in self._factory.prefixes_rev:
					self._full_id = ident
					ident = "%s:%s" % (self._factory.prefixes_rev[pref], rest)

				self.id = ident
			elif ident == "":
				# Allow explicit setting of empty string
				self.id = ""
			elif ident.startswith("auto "):
				# Allow override of auto setting for this resource
				# eg:  model.Person(ident="auto int") to assign an int, rather than the default
				self.id = factory.generate_id(self, auto_type=ident[5:].strip())
			else:
				# Allow for prefixed term that isn't ambiguously a URI
				curied = ident.split(':', 1)
				if len(curied) == 2 and curied[0] in self._factory.prefixes:
					self.id = ident
					self._full_id = self._factory.prefixes[curied[0]] + curied[1]	
				elif not self.__class__._uri_segment:
					self.id = factory.base_url + ident
				else:
					self.id = factory.base_url + self.__class__._uri_segment + "/" + ident
		elif factory.auto_assign_id:
			self.id = factory.generate_id(self)
		else:
			# Not auto assigning, and not submitted = blank node
			self.id = ""

	def _toJSON(self, done, top=None):
		if self._factory.elasticsearch_compatible:
			return {'id': self.id}
		else:
			return self.id

class BaseResource(ExternalResource):
	"""Base class for all resources with classes"""

	_integer_properties = []
	_object_properties = []
	_required_properties = []
	_warn_properties = []
	_classification = ""
	_classhier = []

	def __init__(self, ident=None, label="", value="", content="", **kw):
		"""Initialize BaseObject."""
		super(BaseResource, self).__init__(ident)

		if self._factory.validate_profile and hasattr(self, '_okayToUse'): 
			if not self._okayToUse:
				raise ProfileError("Class '%s' is configured to not be used" % self.__class__._type)
			elif self._okayToUse == 2:
				self._factory.maybe_warn("Class '%s' is configured to warn on use" % self.__class__._type)

		# Set label and value/content
		if label:
			self._label = label
		# this might raise an exception if value is not allowed on the object
		# but easier to do it in the main init than on many generated subclasses

		is_sym = isinstance(self, SymbolicObject)
		is_dim = isinstance(self, Dimension)

		if value and is_dim:
			self.value = value
		elif content and is_sym:
			self.content = content
		elif value and is_sym:
			self.content = value # not the right param, but not ambiguous
		elif content and is_dim:
			self.value = content # ditto
		elif value or content: 
			raise ProfileError("Class '%s' does not hold values" % self.__class__._type)
		# Custom post initialization function for autoconstructed classes
		self._post_init(**kw)

	def __dir__(self):
		d = dir(self.__class__)
		d.extend(self.list_all_props())
		return sorted(d)

	def __eq__(a, b):
		if id(a) == id(b):
			return True
		if not a or not b:
			return False
		if not isinstance(b, BaseResource) or not isinstance(a, BaseResource):
			return False
		ap = a.list_my_props()
		bp = b.list_my_props()
		if ap != bp:
			return False
		for p in ap:
			av = getattr(a, p)
			bv = getattr(b, p)
			if av != bv:
				return False
		return True

	@property
	def type(self):
		for c in self._classhier:
			if c._type:
				return c.__name__

	@type.setter
	def type(self, value):
		raise AttributeError("Must not set 'type' on resources directly")
	
	def set_context(self, value):	
		raise DataError("Must not set the JSON LD context directly", self)

	def _post_init(self, **kw):
		# Expect this to be overridden / replaced
		pass

	def __setattr__(self, which, value):
		"""Attribute setting magic for error checking and resource/literal handling."""

		if which[0] == "_" or not value:
			# _label goes through here, but it would below anyway, as it takes a Literal
			object.__setattr__(self, which, value)			
		else:
			# Allow per-class setters
			if hasattr(self, 'set_%s' % which):
				fn = getattr(self, 'set_%s' % which)
				return fn(value)

			if self._factory.validate_properties or self._factory.validate_profile or self._factory.validate_range:
				ok = self._check_prop(which, value)
			elif isinstance(value, ExternalResource):
				ok = 2
			else:
				ok = 1

			if ok == 2:
				self._set_magic_resource(which, value)
			else:			
				object.__setattr__(self, which, value)				
		 
	def _check_prop(self, which, value):
		val_props = self._factory.validate_properties
		val_profile = self._factory.validate_profile and getattr(self, '_validate_profile', True)
		val_range = self._factory.validate_range
		for c in self._classhier:
			if which in c._all_properties:
				pinfo = c._all_properties[which]

				if val_profile:
					okay = pinfo.profile_okay					
					rdf = pinfo.predicate
					if not okay:
						raise ProfileError("Property '%s' / '%s' is configured to not be used" % (which, rdf), self)
					elif okay == 2:
						self._factory.maybe_warn("Property '%s' / '%s' is configured to warn on use" % (which, rdf))

				if val_range:
					rng = pinfo.range
					if rng is str:					
						return 1
					elif type(value) is BaseResource:
						# Allow direct instances of base resource anywhere
						# this is an override for external URIs
						return 2
					elif isinstance(value, rng):
						return 2
					else:
						raise DataError("Can't set '%s' on resource of type '%s' to '%r'; must be '%s'" % (which, self._type, value, rng), self)
				# Found it, but not validating range and either okay or not validating profile
				return 1
		if val_props:
			raise DataError("Can't set unknown field '%s' on resource of type '%s'" % (which, self._type), self)
		else:
			# Not validating ANYTHING
			return 1

	def _check_reference(self, data):
		"""True if data is a resource or reference to a resource"""
		# "http://..."
		# {"@id": "http://..."}
		# or list of above
		if type(data) in STR_TYPES:
			return data.startswith('http')
		elif type(data) is dict:
			return 'id' in data
		elif isinstance(data, BaseResource):
			return True
		elif type(data) is list:
			for d in data:
				if type(d) in STR_TYPES and not d.startswith('http'):
					return False
				elif type(d) is dict and not 'id' in d:
					return False
			return True
		else:
			self._factory.maybe_warn("expecing a resource, got: %r" % (data))
			return True
				
	def _set_magic_resource(self, which, value, inversed=False):
		"""Set resource property.
		allow: string/object/dict, and magically generate list thereof
		"""

		if self._factory.materialize_inverses or self._factory.process_multiplicity or \
			self._factory.validate_multiplicity:
			inverse = None
			multiple = 1
			for c in self._classhier:
				if which in c._all_properties:
					v = c._all_properties[which]
					multiple = v.multiple_okay
					if v.inverse_property:
						inverse = v.inverse_property
					break

		try:
			current = getattr(self, which)
		except:
			current = None
		if not current:
			object.__setattr__(self, which, value)
		elif type(current) is list:
			# check value not in list
			if self._factory.multiple_instances_per_property == "error" and isinstance(value, BaseResource) and value in current:
				raise DataError("""Cannot add the same resource in the same property more than once:
change factory.multiple_instances_per_property to 'drop' or 'allow'""")
			current.append(value)
		else:
			if self._factory.validate_multiplicity and not multiple:
				raise ProfileError("Cannot append to %s on %s as multiplicity is 1" % (which, self._type))
			nvalue = [current, value]
			object.__setattr__(self, which, nvalue)

		if self._factory.materialize_inverses and not inversed and inverse:
			# set the backwards ref		
			value._set_magic_resource(inverse, self, True)
		if self._factory.process_multiplicity and type(current) is not list and multiple:
			object.__setattr__(self, which, [getattr(self, which)])


	def _toJSON(self, done, top=None):
		"""Serialize as JSON."""
		# If we're already in the graph, return our URI only
		# This should only be called from the factory!

		d = self.__dict__.copy()
		del d['_factory']

		# Can't pass in self as a param
		if top is None:
			top = self

		# id, type, _label is the default.
		if not factory.id_type_label and id(self) in done:
			if self._factory.elasticsearch_compatible:
				return {'id': self.id}
			else:
				return self.id

		# In case of local contexts, not at the root
		# Shouldn't ever happen, but worth testing for
		if 'context' in d:
			d['@context'] = d['context']
			del d['context']

		# Check mandatory properties
		if self._factory.validate_profile:
			for e in self._required_properties:
				if e not in d:
					raise RequirementError("Resource type '%s' requires '%s' to be set" % (self._type, e), self)

		debug = self._factory.debug_level
		if debug.find("warn") > -1:
			for e in self._warn_properties:
				if e not in d:
					msg = "Resource type '%s' should have '%s' set" % (self._type, e)
					self._factory.maybe_warn(msg)

		# Add back context at the top, if set
		if top is self and self._factory.context_uri: 
			d['@context'] = self._factory.context_uri

		if (self._factory.id_type_label and id(self) in done) or (top is not self and not self._embed):
			# limit to only id, type, label
			nd = {}
			nd['id'] = d['id']
			if self.type:
				nd['type'] = self.type
			try:
				nd['_label'] = d['_label']
			except:
				pass
			d = nd
		else:	
			# otherwise, we're about to serialize the resource completely
			done[id(self)] = 1			

		# Need to do in order now to get done correctly ordered
		if self._factory.order_json:
			KOH = self._factory.key_order_hash
			kodflt = self._factory.key_order_default
			kvs = sorted(d.items(), key=lambda x: KOH.get(x[0], kodflt))
		else:
			kvs = list(d.items())

		tbd = []
		for (k, v) in kvs:
			# some _foo might be carried through, eg _label or _comment
			k = self._property_name_map.get(k, k)
			if (not v and v is not False) or (k[0] == "_" and not k in self._factory.underscore_properties):
				del d[k]
			else:
				if isinstance(v, ExternalResource):
					if self._factory.linked_art_boundaries and \
						not self._linked_art_boundary_okay(top, k, v):
						# never follow, so just add to done
						done[id(v)] = 1
					else:
						tbd.append(id(v))
				elif type(v) is list:
					for ni in v:
						if isinstance(ni, ExternalResource):
							if self._factory.linked_art_boundaries and \
								not self._linked_art_boundary_okay(top, k, ni):
								# never follow, so just add to done
								done[id(ni)] = 1							
							else:
								tbd.append(id(ni))
					# For completeness should check list-of-datetime here too
				elif isinstance(v, datetime.datetime):
					# replace with string
					kvs[k] = v.strftime("%Y-%m-%dT%H:%M:%SZ")

		for t in tbd:
			if not t in done:
				done[t] = id(self)
			
		for (k,v) in kvs:
			nk = self._property_name_map.get(k, k)
			if nk != k:
				del d[k]
				d[nk] = v
				k = nk
			if v and (k[0] != "_" and not k in self._factory.underscore_properties):
				if isinstance(v, ExternalResource):
					if done[id(v)] == id(self):
						del done[id(v)]
					d[k] = v._toJSON(done=done, top=top)
				elif type(v) is list:
					newl = []
					uniq = set()
					for ni in v:
						if self._factory.multiple_instances_per_property == "drop":
							if id(ni) in uniq:
								continue
							else:
								uniq.add(id(ni))
						if isinstance(ni, ExternalResource):
							if done[id(ni)] == id(self):
								del done[id(ni)]
							newl.append(ni._toJSON(done=done, top=top))
						else:
							# A number or string
							newl.append(ni)
					d[k] = newl				

		if self._factory.full_names:
			nd = {}
			# @context gets ganked by this renaming
			# so add it back in first.
			if top is self:
				nd['@context'] = self._factory.context_uri

			for (k,v) in d.items():
				# look up the rdf predicate in _properties
				for c in self._classhier:
					if k in c._all_properties:
						nk = c._all_properties[k].predicate
						nd[nk] = v
						break

			# Ensure full version uses basic @type
			if "rdf:type" in nd:
				nd['@type'] = nd['rdf:type']
				del nd['rdf:type']

			# And type gets ganked for overlay classes (Painting)
			# plus for stupidity classes (DestructionActivity)
			# so add this back too
			if not "@type" in nd or not nd['@type']:
				# find class up that has a type and use its name
				for c in reversed(self._classhier):
					if c._type:
						nd['@type'] = c._type

			d = nd
			KOH = self._factory.full_key_order_hash
		else:
			# Use existing programmer-friendly names for classes too
			# find class up that has a type and use its name
			# almost certainly the first one
			if self.type:
				d['type'] = self.type

			if self._factory.pipe_scoped_contexts:
				# XXX TODO This should be configurable not hard coded
				if 'part' in d:
					# Calculate which part
					for c in reversed(self._classhier):
						if 'part' in c._all_properties:
							nk = c._all_properties['part'].predicate
							d['part|%s' % nk]  = d['part']
							del d['part']
							break
				if 'part_of' in d:
					# Calculate which part
					for c in reversed(self._classhier):
						if 'part_of' in c._all_properties:
							nk = c._all_properties['part_of'].predicate
							d['part_of|%s' % nk]  = d['part_of']
							del d['part_of']
							break

		if self._highlight and self._factory.allow_highlight:
			d['_highlight'] = True
		if self._elide and self._factory.allow_elide:
			d['_elide'] = True

		if self._factory.order_json:
			return OrderedDict(sorted(d.items(), key=lambda x: KOH.get(x[0], 1000)))
		else:
			return d

	def _toJSON_fast(self, done, top=None):
		"""Serialize as JSON."""
		# If we're already in the graph, return our URI only
		# This should only be called from the factory!

		# This should ONLY be used in Python3.6+
		# as it relies on a stable order to a regular dict
		# which was not the case before 3.6

		# id, type, _label is the default.
		if not self._factory.id_type_label and id(self) in done:
			return self.id

		# Can't pass in self as a param
		if top is None:
			top = self

		# Add back context at the top, if set
		result = {}
		if top is self and id(self) not in done and self._factory.context_uri: 
			result['@context'] = self._factory.context_uri

		if self.id:
			# Could be a bnode
			result['id'] = self.id
		if self.type:
			# Could be an external reference
			result['type'] = self.type
		try:
			result['_label'] = self._label
		except:
			pass

		# Need only minimal representation of self
		if (self._factory.id_type_label and id(self) in done) or (top is not self and not self._embed):
			# limit to only id, type, label
			return result
		else:	
			# otherwise, we're about to serialize the resource completely
			done[id(self)] = 1			

		d = self.__dict__.copy()
		del d['_factory']
		del d['id']

		# Need to do in order now to get done correctly ordered
		kvs = list(d.items())
		if self._factory.order_json:
			KOH = self._factory.key_order_hash
			kodflt = self._factory.key_order_default
			# in place sort is 2% faster than sorted
			# rather than over an iterable
			# and sorting is 10% of srlz cost
			# and use setdefault to amortize cost of default
			# over multiple calls to the factory's hash
			kvs.sort(key=lambda x: KOH.setdefault(x[0], kodflt))
			
		# tbd vs done is to ensure that in a DAG rather than a tree
		# that children of the current node are given priority for
		# full serialization, rather than deeper descendent nodes
		# that would otherwise be processed first during the (recursive)
		# depth-first traversal.

		# This doesn't catch the pattern A-B-C-D / A-E-D,
		# (D will be under B-C, not under E) as B is processed completely
		# before E.

		# Note: Reimplemented this in a single depth-first loop
		# and it was less than 2% faster. The 2 loops are not that expensive

		tbd = []
		for (k, v) in kvs:
			if k[0] == "_" and not k in self._factory.underscore_properties:
				continue
			k = self._property_name_map.get(k, k)
			if isinstance(v, ExternalResource):
				if self._factory.linked_art_boundaries and \
					not self._linked_art_boundary_okay(top, k, v):
					# never follow, so just add to done
					done[id(v)] = 1
				else:
					tbd.append(id(v))
			elif type(v) is list:
				for ni in v:
					if isinstance(ni, ExternalResource):
						if self._factory.linked_art_boundaries and \
							not self._linked_art_boundary_okay(top, k, ni):
							# never follow, so just add to done
							done[id(ni)] = 1							
						else:
							tbd.append(id(ni))

		for t in tbd:
			if not t in done:
				done[t] = id(self)
			
		# This is already sorted if needed
		for (k,v) in kvs:
			if k[0] == "_" and not k in self._factory.underscore_properties:
				continue
			k = self._property_name_map.get(k, k)
			if not v:
				pass
			elif isinstance(v, ExternalResource):
				if done[id(v)] == id(self):
					del done[id(v)]
				result[k] = v._toJSON_fast(done=done, top=top)
			elif type(v) is list:
				newl = []
				uniq = set()
				for ni in v:
					if self._factory.multiple_instances_per_property == "drop":
						if id(ni) in uniq:
							continue
						else:
							uniq.add(id(ni))
					if isinstance(ni, ExternalResource):
						if done[id(ni)] == id(self):
							del done[id(ni)]
						newl.append(ni._toJSON_fast(done=done, top=top))
					else:
						# A number or string
						newl.append(ni)
				result[k] = newl
			elif isinstance(v, datetime.datetime):
				result[k] = v.strftime("%Y-%m-%dT%H:%M:%SZ")
			else:
				result[k] = v
		return result

	def _linked_art_boundary_okay(self, top, prop, value):
		# Return false to say do not cross this boundary
		# Without replacement, just return True always
		return True

	def list_all_props(self, filter=None, okay=None):
		props = []
		for c in self._classhier:		
			for k,v in c._all_properties.items():
				if not k in props and \
					(not okay or (okay and v.profile_okay)) and \
					(filter is None or isinstance(filter, v.range) or \
						filter is v.range):
					props.append(k)
		props.sort()
		return props

	def list_all_props_with_range(self, filter=None, okay=None):
		props = {}
		for c in self._classhier:		
			for k,v in c._all_properties.items():
				if not k in props and \
					(not okay or (okay and v.profile_okay)) and \
					(filter is None or isinstance(filter, v.range) or \
						filter is v.range):
					props[k] = v.range
		return props

	def list_my_props(self, filter=None):
		d = self.__dict__.copy()		
		props = []
		for (k,v) in d.items():
			if k[0] != "_" or k in self._factory.underscore_properties:
				# real property
				if filter:
					if isinstance(v, filter):
						props.append(k)
					elif isinstance(v, list):
						for i in v:
							if isinstance(i, filter):
								props.append(k)
								break
				else:
					props.append(k)		
		return props

	def allows_multiple(self, propName):
		""" Does propName allow multiple values on this class """
		for c in self._classhier:
			if propName in c._all_properties:
				v = c._all_properties[propName]
				return bool(v.multiple_okay)
		raise DataError("Cannot set '%s' on '%s'" % (propName, self.__class__.__name__))

	def clone(self, minimal=False):
		""" Make a shallow copy of self. If minimal flag is set, only return id, type and _label """
		new = self.__class__(ident=self.id)
		if minimal:
			if self._label:
				new._label = self._label
		else:
			for p in self.list_my_props():
				curr = getattr(self, p)
				if type(curr) == list:
					for v in curr:
						setattr(new, p, v)
				else:
					setattr(new, p, curr)
		return new

# Ontology / Profile manipulation

def override_okay(clss, propName):
	""" set particular property on the class to be okay to use """
	pinfo = clss._all_properties.get(propName, None)
	if pinfo:
		npinfo = PropInfo(pinfo.property, pinfo.predicate, pinfo.range,
			pinfo.inverse_property, pinfo.inverse_predicate, 
			pinfo.multiple_okay, 1)
		clss._all_properties[propName] = npinfo
	else:
		raise DataError("%s does not have a %s property to allow" % 
			(clss.__name__, propName))


# Ensure everything can have id, type, label and description
BaseResource._properties = {'id': {"rdf": "@id", "range": str, "okayToUse": 1}, 
	'type': {"rdf": "rdf:type", "range": str, "rangeStr": "rdfs:Class", "okayToUse": 1}, 
	'_label': {"rdf": "rdfs:label", "range": str, "rangeStr": "xsd:string", "okayToUse": 1}
}
BaseResource._all_properties = {
	'id': PropInfo('id', '@id', str, None, None, 0, 1),
	'type': PropInfo('type', 'rdf:type', str, None, None, 0, 1),
	'_label': PropInfo('_label', 'rdfs:label', str, None, None, 0, 1)
}
BaseResource._classhier = (BaseResource, ExternalResource)

def process_tsv(fn):
	fh = codecs.open(fn, 'r', 'utf-8')
	lines = fh.readlines()[1:] # chomp header line
	fh.close()
	vocabData = {"rdf:Resource": 
		{"props": [], "label": "Resource", "className": "Resource", 
		"subs":[], "desc": "", "class": BaseResource, "okay": 1}}

	for l in lines:
		l = l[:-1] # chomp
		info= l.split('\t')
		name = info[0]	
		if info[1] == "class":
			data = {"subOf": info[5], "label": info[3], 'className': info[2],
				"desc": info[4], "class": None, "props": [], "subs": [], "okay": info[6]}
			vocabData[name] = data
		else:
			# property
			data = {"name": name, "subOf": info[5], "label": info[3], "propName": info[2],
			"desc": info[4], "range": info[7], "inverse": info[8], "okay": info[10], "multiple": info[11]}
			try:
				what = vocabData[info[6]]
			except:
				print("Failed to find class for %s given %s" % (data, info[6]))
				raise
			what["props"].append(data)

			koh = int(info[9])
			if koh != KEY_ORDER_DEFAULT:
				factory.key_order_hash[data['propName']] = koh
				factory.full_key_order_hash[data['name']] = koh

	# invert subclass hierarchy
	for k, v in vocabData.items():
		if k != "rdf:Resource":		
			sub = v['subOf']
			# | separated list
			for s in sub.split('|'):
				if s:
					try:
						vocabData[s]['subs'].append(k)
					except:
						pass
	return vocabData

# Build class heirarchy recursively
def build_class(crmName, parent, vocabData):

	data = vocabData[crmName]
	name = str(data['className'])

	# check to see if we already exist
	# nb globals() here is only this module
	if name in globals():
		c = globals()[name]
		try:	
			c.__bases__ += (parent,)
		except:
			print("MRO FAILURE: %r --> %r + %r" % (c, c.__bases__, parent))
			raise
		return

	c = type(name, (parent,), {'__doc__': data['desc']})
	globals()[name] = c
	data['class'] = c
	if not ":" in crmName:
		c._type = "crm:%s" % crmName
	else:
		c._type = crmName
	c._uri_segment = name
	c._properties = {}
	c._all_properties = {}
	c._okayToUse = int(data['okay'])

	factory._all_classes[name] = c

	# Set up real properties
	for p in data['props']:
		ccname, pvalue = _make_property_def(p)
		c._properties[ccname] = pvalue

	# Build subclasses
	for s in data['subs']:
		build_class(s, c, vocabData)

def _make_property_def(p):
	name = p['name']
	if not ":" in name:
		name = "crm:%s" % name
	rng = p['range']
	ccname = p['propName']
	if p['inverse']:
		i = p['inverse']
		if i[0] == "P":
			invRdf = "crm:%s" % i
		else:
			invRdf = i
	else:
		invRdf = ""
	okay = p['okay']
	if not okay:
		okay = '1'
	okay = int(okay)
	mult = p['multiple']
	if not mult:
		mult = '0'
	mult = int(mult)
	# can't guarantee all classes have been built at this stage :(
	return (ccname, {"rdf": name, 
		"rangeStr": rng,
		"inverseRdf": invRdf,
		"okayToUse": okay,
		"multiple": mult})

def build_classes(fn=None, topClass=None):
	# Default to building our core dataset

	if not fn:
		dd = os.path.join(os.path.dirname(__file__), 'data')
		fn = os.path.join(dd, 'crm_vocab.tsv')
		topClass = 'E1_CRM_Entity'

	vocabData = process_tsv(fn)

	# Everything can have an id, a type, a label, a description
	build_class(topClass, BaseResource, vocabData)

	# And reset definitions now we have class objects
	for v in vocabData.values():
		c = v['class']
		if c is BaseResource:
			continue
		for (name, value) in c._properties.items():
			# recreate with namedtuple
			rngs = value.get('rangeStr', None)
			if not rngs:
				# Precomputed ones like rdf:type
				continue

			inverse = None
			rngd = vocabData.get(value['rangeStr'], None)
			if rngs in ['rdfs:Literal', 'xsd:dateTime', 'xsd:string', 'rdfs:Class']:
					rng = str 
			elif not rngd:
				raise ConfigurationError("Failed to get range for %s property %s - %s" % (c, name, rngs))
			else:
				rng = rngd['class']
				# and add inverse prop name from range
				if rng == None:
					# Uh-oh!
					raise ConfigurationError("Class %s did not get built, range of %s" % (value['rangeStr'], name))
				for (ik, iv) in rng._properties.items():
					if iv['inverseRdf'] == value['rdf']:
						inverse = ik
						break

			c._all_properties[name] = PropInfo(name,
				value['rdf'], rng, inverse,
				value.get('inverseRdf', None),
				value.get('multiple', 1),
				value.get('okayToUse', 0)
			)

	# set all of the classhiers
	for v in vocabData.values():
		c = v['class']
		c._classhier = inspect.getmro(c)[:-1]
		try:
			del c._properties
		except:
			# Never had it set?
			pass

# XXX This should be invoked rather than inline so the module can be loaded
# and a different context used. But for now ...
# Build the factory first, so properties can be added to key_order
factory = CromulentFactory("http://lod.example.org/museum/", context="https://linked.art/ns/v1/linked-art.json")
build_classes()
# Need to then configure the boundary classes after they're created



import unittest 
import sys
import os
import shutil
import json
import pickle
from collections import OrderedDict
from cromulent import model, vocab
from cromulent.model import override_okay


class TestFactorySetup(unittest.TestCase):

	def setUp(self):
		model.factory.base_url = 'http://data.getty.edu/provenance/'
		model.factory.base_dir = 'tests/provenance_base_dir'
		model.factory.default_lang = 'en'
		#model.factory.context_uri = 'http://www.cidoc-crm.org/cidoc-crm/'

	def tearDown(self):
		model.factory.base_url = 'http://lod.example.org/museum/'
		model.factory.log_stream = sys.stderr
		model.factory.debug_level = 'warn'

	def test_base_url(self):
		self.assertEqual(model.factory.base_url, 'http://data.getty.edu/provenance/')

	def test_base_dir(self):
		self.assertEqual(model.factory.base_dir, 'tests/provenance_base_dir')

	def test_default_lang(self):
		self.assertEqual(model.factory.default_lang, 'en')

	def test_set_debug_stream(self):
		strm = open('err_output', 'w')
		model.factory.set_debug_stream(strm)
		self.assertEqual(model.factory.log_stream, strm)

	def test_set_debug(self):
		model.factory.set_debug('error_on_warning')
		self.assertEqual(model.factory.debug_level, 'error_on_warning')
		self.assertRaises(model.ConfigurationError, model.factory.set_debug, 'xxx')
		self.assertRaises(model.MetadataError, model.factory.maybe_warn, "test")

	def test_load_context(self):
		self.assertRaises(model.ConfigurationError, model.factory.load_context, 
			"foo", {"foo":"does_not_exist.txt"})
		model.factory.load_context("foo", {"foo":"tests/test_context.json"})
		self.assertEqual(model.factory.context_json, {"@context":{"id":"@id"}})
		self.assertRaises(model.ConfigurationError, model.factory.load_context, "", {})

	def test_pickle(self):
		model.factory.log_stream = sys.stderr
		srlz = pickle.dumps(model.factory)
		newfac = pickle.loads(srlz)
		self.assertTrue(model.factory.log_stream is newfac.log_stream)



class TestFactorySerialization(unittest.TestCase):

	def setUp(self):
		self.collection = model.InformationObject('collection')
		self.collection._label = "Test Object"

	def test_broken_unicode(self):
		model.factory.debug_level = "error_on_warning"
		try:
			badval = b"\xFF\xFE\x02"
		except:
			badval = "\xFF\xFE\x02"
		badjs = {"_label": badval}
		self.assertRaises(model.MetadataError, model.factory._buildString,
			js=badjs)

	def test_toJSON(self):
		# model.factory.context_uri = 'http://lod.getty.edu/context.json'
		expect = OrderedDict([
			('@context', model.factory.context_uri),
			('id', u'http://lod.example.org/museum/InformationObject/collection'), 
			('type', 'InformationObject'), ('_label', 'Test Object')])
		outj = model.factory.toJSON(self.collection)
		self.assertEqual(expect, outj)

	def test_toJSON_fast(self):
		model.factory.json_serializer = "fast"
		expect = {'@context': model.factory.context_uri, 
			'id': 'http://lod.example.org/museum/InformationObject/collection', 
			'type': 'InformationObject', 
			'_label': 'Test Object'}
		outj = model.factory.toJSON(self.collection)
		self.assertEqual(expect, outj)
		model.factory.json_serializer = "normal"

	def test_toJSON_normal(self):
		expect = OrderedDict([(u'@context', model.factory.context_uri), 
			(u'@id', u'http://lod.example.org/museum/Person/1'), (u'@type', u'crm:E21_Person'),
			('rdfs:label', 'Test Person')])
		model.factory.full_names = True
		p = model.Person("1")
		p._label = "Test Person"
		outj = model.factory.toJSON(p)
		self.assertEqual(expect, outj)
		# reset
		model.factory.full_names = False

	def test_toString(self):
		expect = u'{"@context":"'+model.factory.context_uri+'","id":"http://lod.example.org/museum/InformationObject/collection","type":"InformationObject","_label":"Test Object"}'
		outs = model.factory.toString(self.collection)
		self.assertEqual(expect, outs)

	def test_toString_fast(self):
		# Should only be trusted in python 3
		if sys.version_info.major >= 3 and sys.version_info.minor >= 6:
			expect = u'{"@context":"'+model.factory.context_uri+'","id":"http://lod.example.org/museum/InformationObject/collection","type":"InformationObject","_label":"Test Object"}'
			model.factory.json_serializer = "fast"		
			outs = model.factory.toString(self.collection)
			model.factory.json_serializer = "normal"
			self.assertEqual(expect, outs)
		else:
			print("Skipping toString_fast test in Python 2.x")

	def test_toFile(self):
		self.assertRaises(model.ConfigurationError, model.factory.toFile, self.collection)
		# Test auto filename determination
		model.factory.base_dir = 'tests'
		model.factory.toFile(self.collection)
		self.assertTrue(os.path.isfile('tests/InformationObject/collection.json'))
		# Test explicit filename setting
		model.factory.toFile(self.collection, filename='tests/fishbat.bar')
		self.assertTrue(os.path.isfile('tests/fishbat.bar'))
		# Tidy up
		shutil.rmtree('tests/InformationObject')

	def test_breadth(self):
		x = model.TransferOfCustody()
		e = model.Activity()
		fr = model.Group()
		to = model.Group()
		w = model.HumanMadeObject()
		fr._label = "From"
		to._label = "To"
		x.transferred_custody_of = w
		x.transferred_custody_from = fr
		x.transferred_custody_to = to
		e.used_specific_object = w
		e.carried_out_by = to
		w.current_owner = fr
		x.specific_purpose = e
		js = model.factory.toJSON(x)
		# Okay ... if we're breadth first, then custody_from is a resource
		# And now it's the first in the list
		self.assertTrue(isinstance(js['transferred_custody_from'][0], OrderedDict))

	def test_string_list(self):
		x = model.Activity()
		x._label = ["Label 1", "Label 2"]
		js = model.factory.toJSON(x)
		self.assertTrue(js['_label'] == x._label)

	def test_external(self):
		x = model.ExternalResource(ident="1")
		model.factory.elasticsearch_compatible = 1
		js = x._toJSON(done=None)
		self.assertTrue(type(js) == dict)
		model.factory.elasticsearch_compatible = 0
		js = x._toJSON(done=None)
		# testing unicode in 2, str in 3 :(
		self.assertTrue(type(js) != dict)		

	def test_recursion(self):
		x = model.Activity()
		x.part = x
		js = model.factory.toJSON(x)
		# If our recursion checks have regressed, this will barf right here
		self.assertTrue(1)

	def test_pipe_scoped(self):
		x = model.Activity()
		y = model.Activity()
		x.part = y
		model.factory.pipe_scoped_contexts = True
		js = model.factory.toJSON(x)
		self.assertTrue('part|crm:P9_consists_of' in js)
		model.factory.pipe_scoped_contexts = False
		js = model.factory.toJSON(x)		
		self.assertTrue('part|crm:P9_consists_of' not in js)		
		self.assertTrue('part' in js)

	def test_collapse_json(self):
		model.factory.auto_id_type = "uuid"
		model.factory.base_url = "http://lod.example.org/museum/"
		model.factory.context_uri = "https://linked.art/ns/v1/linked-art.json"
		p = model.Person()
		p.classified_as = model.Type(ident="http://example.org/Type", label="Test")
		res1 = model.factory.toString(p, compact=False, collapse=60) # all new lines
		res2 = model.factory.toString(p, compact=False, collapse=120) # compact list of type
		self.assertEqual(len(res1.splitlines()), 12)
		self.assertEqual(len(res2.splitlines()), 6)

	def test_production_mode(self):

		# model.factory.production_mode()
		# Can't unset the cached hierarchy
		# and it causes the test for the hierarchy to fail
		model.factory.validate_profile = False
		model.factory.validate_properties = False
		model.factory.validate_range = False
		model.factory.validate_multiplicity = False

		p = model.Person()
		p.identified_by = model.Name(value="abc")
		p.part = model.HumanMadeObject()
		js = model.factory.toJSON(p)

		model.factory.production_mode(state=False)


	def test_ordering(self):
		p = model.Person(label="Person")
		p.classified_as = model.Type(ident="type-uri")
		p.referred_to_by = model.LinguisticObject(content="text")
		p.dimension = model.Dimension(value=1)

		outstr = model.factory.toString(p)
		lbl = outstr.index("_label")
		clsf = outstr.index("classified_as")
		r2b = outstr.index("referred_to_by")
		dim = outstr.index("dimension")
		self.assertTrue(lbl < clsf)
		self.assertTrue(clsf < r2b)
		self.assertTrue(r2b < dim)


class TestProcessTSV(unittest.TestCase):

	def test_process_tsv(self):
		expect = {u'subs': [u'E84_Information_Carrier'], u'label': u'Human-Made Object', u'className': u'HumanMadeObject', 
		u'subOf': u'E19_Physical_Object|E24_Physical_Human-Made_Thing', u'props': [], u'class': None, u'okay': u'1'}
		fn = 'cromulent/data/crm_vocab.tsv'
		vocabData = model.process_tsv(fn)
		man_made = vocabData['E22_Human-Made_Object']
		del man_made['desc']  # too long and volatile
		# check subs specifically - could be coming from an extension
		if man_made['subs'] != expect['subs']:
			del man_made['subs']
			del expect['subs']
		self.assertEqual(expect, man_made)


class TestBuildClasses(unittest.TestCase):

	def test_build_classes(self):
		tsv = "\nClassName_full\tclass\tClassName_py\tClass Label\tClass Description\t\t1\t\n"
		fh = open('tests/temp.tsv', 'w')
		fh.write(tsv)
		fh.close()
		model.build_classes("tests/temp.tsv", "ClassName_full")
		from cromulent.model import ClassName_py
		self.assertEqual('Class Description', ClassName_py.__doc__)
		os.remove('tests/temp.tsv')

class TestBuildClass(unittest.TestCase):

	def test_build_class(self):
		tsv = "\nClassName_full\tclass\tClassName_py2\tClass Label\tClass Description\t\t1\t\n"
		fh = open('tests/temp.tsv', 'w')
		fh.write(tsv)
		fh.close()
		vocabData = model.process_tsv('tests/temp.tsv')
		model.build_class('ClassName_full', model.BaseResource, vocabData)
		from cromulent.model import ClassName_py2
		self.assertEqual('Class Description', ClassName_py2.__doc__)
		os.remove('tests/temp.tsv')

class TestAutoIdentifiers(unittest.TestCase):

	def test_bad_autoid(self):
		model.factory.auto_assign_id = True
		model.factory.auto_id_type = "broken"
		self.assertRaises(model.ConfigurationError, model.factory.generate_id,
			"irrelevant")

	def test_int(self):
		model.factory.auto_assign_id = True
		model.factory.auto_id_type = "int"
		p = model.Person()
		p2 = model.Activity()
		self.assertEqual(int(p.id[-1]), int(p2.id[-1])-1)

	def test_int_per_type(self):
		model.factory.auto_assign_id = True
		model.factory.auto_id_type = "int-per-type"
		p = model.Person()
		p2 = model.Person()
		self.assertEqual(int(p.id[-1]), int(p2.id[-1])-1)
		p3 = model.Activity()
		self.assertEqual(int(p.id[-1]), int(p3.id[-1]))		

	def test_int_per_segment(self):
		model.factory.auto_assign_id = True
		model.factory._auto_id_segments = {}
		model.factory.auto_id_type = "int-per-segment"
		model.Activity._uri_segment = model.Person._uri_segment
		p = model.Person()
		p2 = model.Activity()
		self.assertEqual(int(p.id[-1]), int(p2.id[-1])-1)		
		p3 = model.TimeSpan()
		self.assertEqual(int(p.id[-1]), int(p3.id[-1]))		
			
	def test_uuid(self):
		model.factory.auto_assign_id = True
		model.factory.auto_id_type = "uuid"
		p = model.Person()
		self.assertTrue(p.id.startswith('urn:uuid:'))		

	def test_prefixes(self):

		model.factory.prefixes = {'fish':'http://example.org/ns/'}
		p3 = model.Person('fish:3')
		self.assertEqual(p3.id, 'fish:3')
		self.assertEqual(p3._full_id, 'http://example.org/ns/3')

		model.factory.prefixes = {}
		p4 = model.Person('fish:4')
		self.assertTrue(p4.id.startswith(model.factory.base_url))

	def test_other_uris(self):
		p1 = model.Person(ident="tag:some-info-about-person")
		self.assertEqual(p1.id, "tag:some-info-about-person")
		p2 = model.Person(ident="info:ulan/500012345")
		self.assertEqual(p2.id, "info:ulan/500012345")
		p3 = model.Person(ident="some:random:thing:with:colons")
		self.assertFalse(p3.id == "some:random:thing:with:colons")

	def test_no_ident(self):

		model.factory.auto_assign_id = True
		p1 = model.Person()	# auto assigned	 
		p2 = model.Person(ident=None) # auto assigned
		p3 = model.Person(ident="") # bnode explicitly

		self.assertTrue(p1.id.startswith('http'))
		self.assertTrue(p2.id.startswith('http'))
		self.assertEqual(p3.id, '')

		model.factory.auto_assign_id = False
		p4 = model.Person() # bnode is default
		p5 = model.Person(ident=None) # bnode is default
		p6 = model.Person(ident="") # bnode explicitly

		self.assertEqual(p4.id, '')
		self.assertEqual(p5.id, '')
		self.assertEqual(p6.id, '')

		
class TestBaseResource(unittest.TestCase):

	def setUp(self):
		override_okay(model.Person, 'parent_of')
		self.artist = model.Person('00001', 'Jane Doe')
		self.son = model.Person('00002', 'John Doe')

	def test_init(self):
		self.assertEqual(self.artist.id, 'http://lod.example.org/museum/Person/00001')
		self.assertEqual(self.artist._type, 'crm:E21_Person')
		self.assertEqual(self.artist.type, 'Person')
		self.assertEqual(self.artist._label, 'Jane Doe')
		self.assertFalse(hasattr(self.artist, 'value'))
		self.assertFalse(hasattr(self.artist, 'has_type'))

	def test_check_prop(self):
		desc = self.artist._check_prop('_label', 'Jane Doe\'s Bio')
		self.assertEqual(desc, 1)
		parent = self.artist._check_prop('parent_of', self.son)
		self.assertEqual(parent, 2)

	def test_list_all_props(self):
		props = self.artist.list_all_props()
		props.sort()
		self.assertEqual(props[-1], 'witnessed')
		self.assertTrue('_label' in props)
		self.assertTrue('identified_by' in props)

	def test_list_my_props(self):
		p1 = model.Person()
		p1.classified_as = model.Type()
		props = p1.list_my_props()
		self.assertEqual(set(props), set(['classified_as', 'id']))
		props = p1.list_my_props(filter=model.Type)
		self.assertEqual(props, ['classified_as'])

	def test_allows_multiple(self):
		p = model.Person()
		self.assertTrue(p.allows_multiple('classified_as'))
		self.assertFalse(p.allows_multiple('born'))
		self.assertRaises(model.DataError, p.allows_multiple, 'fish')

	def test_check_reference(self):
		self.assertTrue(self.artist._check_reference('http'))
		self.assertFalse(self.artist._check_reference('xxx'))
		self.assertTrue(self.artist._check_reference({'id': 'xxx'}))
		self.assertFalse(self.artist._check_reference({'xxx': 'yyy'}))
		self.assertTrue(self.artist._check_reference(self.son))
		self.assertTrue(self.artist._check_reference(['http']))
		self.assertFalse(self.artist._check_reference(['xxx', 'yyy']))
		self.assertTrue(self.artist._check_reference(model.Person))

	def test_multiplicity(self):
		model.factory.process_multiplicity = True
		who = model.Actor()
		mmo = model.HumanMadeObject()
		prod = model.Production()
		mmo.produced_by = prod
		who.current_owner_of = mmo
		mmo.current_owner = who
		self.assertEqual(mmo.current_owner, [who])
		self.assertEqual(who.current_owner_of, [mmo])		
		self.assertEqual(mmo.produced_by, prod)

	def test_init_params(self):
		p1 = model.Person(ident="urn:uuid:1234")
		self.assertEqual(p1.id, "urn:uuid:1234")
		p2 = model.Person(ident="http://schema.org/Foo")
		self.assertEqual(p2.id, "schema:Foo")
		p3 = model.Name(content="Test")
		self.assertEqual(p3.content, "Test")
		c = model.MonetaryAmount(value=10)
		self.assertEqual(c.value, 10)
		n = model.Name(value="Rob")
		self.assertEqual(n.content, "Rob")
		i = model.Identifier(content="xyz123")
		self.assertEqual(i.content, "xyz123")
		i2 = model.Identifier(value="abc")
		self.assertEqual(i2.content, "abc")

	def test_dir(self):
		props = dir(self.artist)
		self.assertTrue('identified_by' in props)

	def test_fullNames(self):
		p1 = model.Person()
		p1.identified_by = model.Name(content="Test")
		g1 = model.Group()
		p1.member_of = g1

		model.factory.full_names = True
		js = model.factory.toJSON(p1)
		self.assertTrue('crm:P107i_is_current_or_former_member_of' in list(js.keys()))
		self.assertTrue('crm:P1_is_identified_by' in list(js.keys()))

		i1 = model.HumanMadeObject()
		s1 = model.Set()
		i1.member_of = s1
		js = model.factory.toJSON(i1)
		self.assertTrue('la:member_of' in list(js.keys()))

		model.factory.full_names = False




class TestPropertyCache(unittest.TestCase):

	def test_cache_hierarchy(self):
		o = model.HumanMadeObject()
		self.assertEqual(o._all_properties, {})
		model.factory.cache_hierarchy()
		self.assertTrue(len(o._all_properties) > 50)
		

class TestMagicMethods(unittest.TestCase):

	def setUp(self):
		override_okay(model.Person, 'parent_of')
		# model.Person._properties['parent_of']['multiple'] = 1

	def test_set_magic_resource(self):
		artist = model.Person('00001', 'Jane Doe')
		son = model.Person('00002', 'John Doe')
		daughter = model.Person('00002', 'Jenny Doe')
		son2 = model.Person('00002', 'Jim Doe')
		artist._set_magic_resource('parent_of', son)
		self.assertEqual(artist.parent_of, [son])
		artist._set_magic_resource('parent_of', daughter)
		try:
			self.assertIn(son, artist.parent_of)
			self.assertIn(daughter, artist.parent_of)
		except:
			# 2.6 doesn't have assertIn
			self.assertTrue(son in artist.parent_of)
			self.assertTrue(daughter in artist.parent_of)

		artist._set_magic_resource('parent_of', son2)
		try:
			self.assertIn(son, artist.parent_of)
			self.assertIn(daughter, artist.parent_of)
			self.assertIn(son2, artist.parent_of)
		except:
			self.assertTrue(son in artist.parent_of)
			self.assertTrue(daughter in artist.parent_of)
			self.assertTrue(son2 in artist.parent_of)

	def test_set_magic_resource_inverse(self):
		model.factory.materialize_inverses = True
		artist = model.Person('00001', 'Jane Doe')
		son = model.Person('00002', 'John Doe')
		artist._set_magic_resource('parent_of', son)
		self.assertEqual(son.parent, [artist])
		model.factory.materialize_inverses = False

	def test_validate_profile_off(self):
		model.factory.validate_profile = False
		ia = model.IdentifierAssignment()
		# If it's not turned off this should raise
		model.factory.validate_profile = True
		self.assertRaises(model.ProfileError, model.IdentifierAssignment)		
		p1 = model.Person()
		self.assertRaises(model.ProfileError, p1.__setattr__, 'documented_in', "foo")

	def test_validation_unknown(self):
		model.factory.validate_properties = True
		artist = model.Person('00001', 'Jane Doe')		
		self.assertRaises(model.DataError, artist.__setattr__, 'unknown_property', 1)

	def test_validation_wrong_type(self):
		model.factory.validate_properties = True
		artist = model.Person('00001', 'Jane Doe')	
		self.assertRaises(model.DataError, artist.__setattr__, 'parent_of', 'Bad Value')

	def test_validation_off(self):
		model.factory.validate_properties = False
		artist = model.Person('00001', 'Jane Doe')		
		artist.unknown_property = 1
		self.assertEqual(artist.unknown_property, 1)
		model.factory.validate_properties = True

	def test_validate_multiplicity(self):
		model.factory.validate_multiplicity = True
		who = model.Person()
		b1 = model.Birth()
		who.born = b1
		b2 = model.Birth()
		self.assertRaises(model.ProfileError, who.__setattr__, 'born', b2)
		model.factory.validate_multiplicity = False
		who.born = b2
		self.assertEqual(who.born, [b1, b2])

	def test_not_multiple_instance(self):
		who = model.Person()
		n = model.Name(content="Test")
		who.identified_by = n

		model.factory.multiple_instances_per_property = "error"
		self.assertRaises(model.DataError, who.__setattr__, 'identified_by', n)
		self.assertEqual(who.identified_by, [n])

		model.factory.multiple_instances_per_property = "drop"
		who.identified_by = n
		self.assertEqual(who.identified_by, [n,n])		
		# and check that only serialized once
		js = model.factory.toJSON(who)
		self.assertEqual(len(js['identified_by']), 1)

		model.factory.multiple_instances_per_property = "allow"
		js = model.factory.toJSON(who)
		self.assertEqual(len(js['identified_by']), 2)


class TestObjectEquality(unittest.TestCase):
	def setUp(self):
		self.artist = model.Person('00001', 'Jane Doe')
		self.son = model.Person('00002', 'John Doe')
		self.daughter = model.Person('00002', 'Jenny Doe')
		self.son2 = model.Person('00002', 'Jim Doe')

	def test_eq_ident(self):
		self.assertEqual(self.artist, self.artist)
		self.assertEqual(self.son, model.Person('00002', 'John Doe'))
		self.assertEqual(self.son2, model.Person('00002', 'Jim Doe'))
		self.assertEqual(self.daughter, model.Person('00002', 'Jenny Doe'))

	def test_eq_value(self):
		self.assertEqual(self.artist, model.Person('00001', 'Jane Doe'))
		self.assertEqual(self.son, self.son)
		self.assertEqual(self.son2, self.son2)
		self.assertEqual(self.daughter, self.daughter)

	def test_in_value(self):
		people = (
			model.Person('00001', 'Jane Doe'), # artist
			model.Person('00002', 'Jim Doe')   # son2
		)
		self.assertIn(self.artist, people)
		self.assertNotIn(self.son, people)
		self.assertNotIn(self.daughter, people)
		self.assertIn(self.son2, people)

	def test_neq(self):
		self.assertNotEqual(self.artist, self.son)
		self.assertNotEqual(self.artist, model.Person('00001', 'Jane')) # label differs
		self.assertNotEqual(self.artist, self.daughter)
		self.assertNotEqual(self.artist, self.son2)
		self.assertNotEqual(self.son, self.daughter)
		self.assertNotEqual(self.son, self.son2)
		self.assertNotEqual(self.daughter, self.son2)

	def nation(self, name, ident):
		place = vocab.Place(ident='tag:getty.edu,2019:digital:pipeline:provenance:REPLACE-WITH-UUID#PLACE-COUNTRY-' + name, label=name)
		nation = model.Place(ident=ident)
		place.classified_as = vocab.instances['nation']
		place.identified_by = model.Name(ident='', content=name)
		return place

	def test_equality(self):
		from cromulent.model import factory
		place1 = self.nation('Belgium', 'http://vocab.getty.edu/aat/300128207')
		place2 = self.nation('Belgium', 'http://vocab.getty.edu/aat/300128207')
		self.assertEqual(place1, place2)

if __name__ == '__main__':
	unittest.main()





import unittest 

try:
	from collections import OrderedDict
except:
	# 2.6
	from ordereddict import OrderedDict

from cromulent import model, vocab
from cromulent.model import factory

class TestAddClassification(unittest.TestCase):
	def test_add_classification(self):
		amnt = model.MonetaryAmount(ident='')
		amnt.value = 7.0
		self.assertNotIn('Asking Price', factory.toString(amnt))
		vocab.add_classification(amnt, vocab.AskingPrice)
		self.assertIn('Asking Price', factory.toString(amnt))

if __name__ == '__main__':
	unittest.main()



#!/usr/bin/env python3

import unittest
try:
	from contextlib import suppress
except:
	# Python 2.7
	suppress = None
import pprint
from datetime import datetime
from cromulent import model, vocab
from cromulent.extract import extract_monetary_amount
import cromulent.extract

CUSTOM_MAPPING = {
	'xxx': vocab.register_instance('xxx custom currency', {'parent': model.Currency, 'id': '999999999', 'label': 'My Dollars'}),
	'zzz': 'us dollars'
}

class TestCurrencyExtraction(unittest.TestCase):
	'''
	Test the ability to extract currency data.
	'''
	def setUp(self):
		pass

	def tearDown(self):
		pass

	def test_extract_simple(self):
		e = extract_monetary_amount({
			'price': '10.0',
			'currency': 'pounds'
		})
		self.assertEqual(e.type, 'MonetaryAmount')
		self.assertEqual(e._label, '10.00 pounds')
		self.assertEqual(e.value, 10)
		c = e.currency
		self.assertEqual(c.type, 'Currency')
		self.assertEqual(c._label, 'British Pounds')

	def test_extract_comma_separated(self):
		e = extract_monetary_amount({
			'price': '1,280.5',
			'currency': 'pounds'
		})
		self.assertEqual(e.type, 'MonetaryAmount')
		self.assertEqual(e._label, '1,280.50 pounds')
		self.assertEqual(e.value, 1280.50)
		c = e.currency
		self.assertEqual(c.type, 'Currency')
		self.assertEqual(c._label, 'British Pounds')

	def test_extract_label_digits(self):
		e = extract_monetary_amount({
			'price': '1,280.5',
			'currency': 'pounds'
		}, truncate_label_digits=4)
		self.assertEqual(e.type, 'MonetaryAmount')
		self.assertEqual(e._label, '1,280.5000 pounds')
		self.assertEqual(e.value, 1280.50)
		c = e.currency
		self.assertEqual(c.type, 'Currency')
		self.assertEqual(c._label, 'British Pounds')

	def test_extract_multiple_comma_separated(self):
		e = extract_monetary_amount({
			'price': '1,310,720.5',
			'currency': 'pounds'
		})
		self.assertEqual(e.type, 'MonetaryAmount')
		self.assertEqual(e._label, '1,310,720.50 pounds')
		self.assertEqual(e.value, 1310720.5)
		c = e.currency
		self.assertEqual(c.type, 'Currency')
		self.assertEqual(c._label, 'British Pounds')

	def test_extract_est(self):
		e = extract_monetary_amount({
			'est_price': '12.0',
			'currency': 'pounds'
		})
		self.assertEqual(e.value, 12)
		c = e.currency
		self.assertEqual(e.classified_as[0]._label, 'Estimated Price')
		self.assertEqual(e.currency._label, 'British Pounds')

	def test_extract_start(self):
		e = extract_monetary_amount({
			'start_price': '8.5',
			'currency': 'pounds'
		})
		self.assertEqual(e.value, 8.5)
		c = e.currency
		self.assertEqual(e.classified_as[0]._label, 'Starting Price')
		self.assertEqual(e.currency._label, 'British Pounds')

	def test_extract_custom_currency_key(self):
		d = {
			'price': '7',
			'currency': 'zzz'
		}
		with self.assertRaises(AttributeError):
			e = extract_monetary_amount(d)
			self.assertEqual(e.currency._label, 'Custom Currency')
		
		e = extract_monetary_amount(d, currency_mapping=CUSTOM_MAPPING)
		self.assertEqual(e.value, 7)
		self.assertEqual(e.currency._label, 'US Dollars')

	def test_extract_custom_currency_instance(self):
		d = {
			'price': '7',
			'currency': 'xxx'
		}
		with self.assertRaises(AttributeError):
			e = extract_monetary_amount(d)
			self.assertEqual(e.currency._label, 'Custom Currency')
		
		e = extract_monetary_amount(d, currency_mapping=CUSTOM_MAPPING)
		self.assertEqual(e.value, 7)
		self.assertEqual(e.currency._label, 'My Dollars')

	def test_extract_price_with_citation(self):
		d = {
			'price': '7',
			'currency': 'pounds',
			'citation': 'crom test suite'
		}
		e = extract_monetary_amount(d, add_citations=True)
		self.assertEqual(e.value, 7)
		self.assertEqual(e.currency._label, 'British Pounds')
		self.assertEqual(e.referred_to_by[0].content, 'crom test suite')


if __name__ == '__main__':
	unittest.main()



#!/usr/bin/env python3

import unittest
try:
	from contextlib import suppress
except:
	# Python 2.7
	suppress = None
import pprint
from datetime import datetime
from cromulent.extract import Dimension, normalized_dimension_object
import cromulent.extract

class TestDimensionExtraction(unittest.TestCase):
	'''
	Test the ability to extract various formats of dimensions.
	'''
	def setUp(self):
		pass

	def tearDown(self):
		pass

	def test_parse_simple_dimensions(self):
		'''
		Test the documented formats that `cromulent.extract.parse_simple_dimensions` can parse
		and ensure that it returns the expected data.
		'''
		tests = {
			"3'": [Dimension(3, 'feet', None)],
			'3 feet': [Dimension(3, 'feet', None)],
			'3 foot': [Dimension(3, 'feet', None)],
			'3 ft': [Dimension(3, 'feet', None)],
			'3 ft.': [Dimension(3, 'feet', None)],
			'2"': [Dimension(2, 'inches', None)],
			'2 in': [Dimension(2, 'inches', None)],
			'2 in.': [Dimension(2, 'inches', None)],
			'2 inch': [Dimension(2, 'inches', None)],
			'2 inches': [Dimension(2, 'inches', None)],
			'2 duymen': [Dimension(2, 'inches', None)],
			'2 d.': [Dimension(2, 'inches', None)],
			'2 d': [Dimension(2, 'inches', None)],
			'''2'8"''': [Dimension(2, 'feet', None), Dimension(8, 'inches', None)],
			'4cm': [Dimension(4, 'cm', None)],
			'2 pieds 3 pouces': [Dimension(2, 'fr_feet', None), Dimension(3, 'fr_inches', None)],
			'1 pied 7 pouces': [Dimension(1, 'fr_feet', None), Dimension(7, 'fr_inches', None)],
			'8 pouce': [Dimension(8, 'fr_inches', None)],
			'8 pouces': [Dimension(8, 'fr_inches', None)],
			'8 1/2 pouces': [Dimension(8.5, 'fr_inches', None)],
			'8 1/4 pouces': [Dimension(8.25, 'fr_inches', None)],
			'8 1/8 pouces': [Dimension(8.125, 'fr_inches', None)],
			'1': [Dimension(1, None, None)],
			
			# values without a unit that follow values with a unit stay in the same system but using the next-finer unit
			'2 pieds 3': [Dimension(2, 'fr_feet', None), Dimension(3, 'fr_inches', None)],
			"1' 3": [Dimension(1, 'feet', None), Dimension(3, 'inches', None)],
		}

		for value, expected in tests.items():
			dims = cromulent.extract.parse_simple_dimensions(value)
			if expected is not None:
				self.assertIsInstance(dims, list)
				self.assertEqual(dims, expected, msg='dimensions: %r' % (value,))
			else:
				self.assertIsNone(dims)

	def test_dimension_cleaner(self):
		'''
		Test the documented formats that `cromulent.extract.dimensions_cleaner` can parse
		and ensure that it returns the expected data.
		'''
		tests = {
			'''2 in by 1 in''': ([Dimension(2, 'inches', None)], [Dimension(1, 'inches', None)]),
			'''2'2"h x 2'8"w''': ([Dimension(2, 'feet', 'height'), Dimension(2, 'inches', 'height')], [Dimension(2, 'feet', 'width'), Dimension(8, 'inches', 'width')]),
			'''1'3"x4cm h''': ([Dimension(1, 'feet', None), Dimension(3, 'inches', None)], [Dimension(4, 'cm', 'height')]),
			'''1'3" by 4"''': ([Dimension(1, 'feet', None), Dimension(3, 'inches', None)], [Dimension(4, 'inches', None)]),
			'Haut 14 pouces, large 10 pouces': ([Dimension(14, 'fr_inches', 'height')], [Dimension(10, 'fr_inches', 'width')]),
			'Haut. 48 pouces, large 68 pouces': ([Dimension(48, 'fr_inches', 'height')], [Dimension(68, 'fr_inches', 'width')]),
			'1 by 4': ([Dimension(1, None, None)], [Dimension(4, None, None)]),
			'Hoog. 6 v., breed 3 v': ([Dimension(6, 'feet', 'height')], [Dimension(3, 'feet', 'width')]),
			'Breedt 6 v., hoog 3 v': ([Dimension(6, 'feet', 'width')], [Dimension(3, 'feet', 'height')]),
			'20 cm x 24,5 cm': ([Dimension(20, 'cm', None)], [Dimension(24.5, 'cm', None)]),
			'2 w by 5 h': ([Dimension(2, None, 'width')], [Dimension(5, None, 'height')]),
			'Hauteur 1 pied 4 pouces, largeur 1 pied 1/2 pouc.': ([Dimension(1, 'fr_feet', 'height'), Dimension(value=4, unit='fr_inches', which='height')], [Dimension(1, 'fr_feet', 'width'), Dimension(value=0.5, unit='fr_inches', which='width')]),
			'h.73 pouces 4 lignes, l.50 pouces': ([Dimension(value=73, unit='fr_inches', which='height'), Dimension(value=4, unit='ligne', which='height')], [Dimension(value=50, unit='fr_inches', which='width')]),
			'haut. 5 pouc. larg. 5 pouc. 4 linges': ([Dimension(value=5, unit='fr_inches', which='height')], [Dimension(value=5, unit='fr_inches', which='width'), Dimension(value=4, unit='ligne', which='width')]),
			'haut. 9 pouc. 4 lignes larg. 10 pouc. 4 linges': ([Dimension(value=9, unit='fr_inches', which='height'), Dimension(value=4, unit='ligne', which='height')], [Dimension(value=10, unit='fr_inches', which='width'), Dimension(value=4, unit='ligne', which='width')]),
			'h 38 cm, w 27 cm': ([Dimension(38, 'cm', 'height')], [Dimension(27, 'cm', 'width')]),
			"hauteur 9 pouces, largeur 7": ([Dimension(value=9, unit='fr_inches', which='height')], [Dimension(value=7, unit=None, which='width')]),
		}

		for value, expected in tests.items():
			dims = cromulent.extract.dimensions_cleaner(value)
			if expected is not None:
				self.assertIsInstance(dims, tuple)
# 				print('===== got:')
# 				pprint.pprint(dims)
# 				print('----- expected:')
# 				pprint.pprint(expected)
# 				print('=====')
				self.assertEqual(dims, expected, msg='dimensions: %r' % (value,))
			else:
				self.assertIsNone(dims)

	def test_extract_physical_dimensions(self):
		'''
		Test the documented formats that `cromulent.extract.extract_physical_dimensions`
		can parse and ensure that it returns the expected data.
		'''
		tests = {}
		h9l7_height = cromulent.vocab.Height(ident='', content=9.0)
		h9l7_height.identified_by = cromulent.model.Name(ident='', content='9 French inches')
		h9l7_height.unit = cromulent.vocab.instances.get('fr_inches')
		h9l7_width = cromulent.vocab.Width(ident='', content=7.0)
		tests["hauteur 9 pouces, largeur 7"] = [h9l7_height, h9l7_width]

		for value, expected_dims in tests.items():
			dims = list(cromulent.extract.extract_physical_dimensions(value))
			for got, expected in zip(dims, expected_dims):
				self.assertEqual(got.value, expected.value)
				self.assertEqual(got.type, expected.type)

				if suppress is None:
					# Python 2.7
					if hasattr(expected, 'unit'):
						self.assertEqual(got.unit, expected.unit)						
					if hasattr(expected, 'classified_as'):
						self.assertEqual(got.classified_as, expected.classified_as)						
					if hasattr(expected, 'identified_by'):
						self.assertEqual(got.identified_by, expected.identified_by)						
				else:			
					with suppress(AttributeError):
						self.assertEqual(got.unit, expected.unit)
					with suppress(AttributeError):
						self.assertEqual(got.classified_as, expected.classified_as)
					with suppress(AttributeError):
						self.assertEqual(got.identified_by, expected.identified_by)

	def test_extract_physical_dimensions_with_default(self):
		'''
		Test the documented formats that `cromulent.extract.extract_physical_dimensions`
		can parse, specifying a default unit, and ensure that it returns the expected data.
		'''
		tests = {}
		h9l7_height = cromulent.vocab.Height(ident='', content=9.0)
		h9l7_height.identified_by = cromulent.model.Name(ident='', content='9 French inches')
		h9l7_height.unit = cromulent.vocab.instances.get('fr_inches')
		h9l7_width = cromulent.vocab.Width(ident='', content=7.0)
		h9l7_width.unit = cromulent.vocab.instances.get('inches')
		tests["hauteur 9 pouces, largeur 7"] = [h9l7_height, h9l7_width]

		for value, expected_dims in tests.items():
			dims = list(cromulent.extract.extract_physical_dimensions(value, default_unit='inches'))
			for got, expected in zip(dims, expected_dims):
				self.assertEqual(got.value, expected.value)
				self.assertEqual(got.type, expected.type)
				self.assertEqual(got.unit, expected.unit)

	def test_normalize_dimension(self):
		tests = {
			'1 ft, 2 in': ('1 foot, 2 inches', Dimension(value=14, unit='inches', which=None)),
			'8 1/2 pouces': ('8.5 French inches', Dimension(value=8.5, unit='fr_inches', which=None)),
			'1 pied 7 pouces': ('1 French foot, 7 French inches', Dimension(value=19, unit='fr_inches', which=None)),
			'2 pied 1/2 pouces': ('2 French feet, 0.5 French inches', Dimension(value=24.5, unit='fr_inches', which=None)),
			'1 pied 3 pouce. 3 linges': ('1 French foot, 3 French inches, 3 lignes', Dimension(value=15.25, unit='fr_inches', which=None)),
			"4' 8": ('4 feet, 8 inches', Dimension(value=56, unit='inches', which=None)),
			"1 pied 2": ('1 French foot, 2 French inches', Dimension(value=14, unit='fr_inches', which=None)),
		}
		for value, expected in tests.items():
			elabel, edim = expected
			dims = cromulent.extract.parse_simple_dimensions(value)
			dim, label = normalized_dimension_object(dims)
			self.assertEqual(label, elabel)
			self.assertEqual(dim, edim)

if __name__ == '__main__':
	unittest.main()



# -*- coding: utf-8 -*-

import re
import warnings
from collections import namedtuple
from contextlib import contextmanager

import locale
import calendar
from contextlib import contextmanager, suppress
from datetime import datetime, timedelta

from cromulent import model, vocab

#mark - Mapping Dictionaries

# TODO: can this be refactored somewhere?
CURRENCY_MAPPING = {
	'österreichische schilling': 'at shillings',
	'florins': 'de florins',
	'fl': 'de florins',
	'fl.': 'de florins',
	'pounds': 'gb pounds',
	'livres': 'fr livres',
	'guineas': 'gb guineas',
	'reichsmark': 'de reichsmarks'
}

#mark - Dimensions

NEXT_FINER_DIMENSION_UNIT = {
	'inches': None,
	'feet': 'inches',
	'cm': None,
	'fr_feet': 'fr_inches',
	'fr_inches': 'ligne'
}
NUMBER_PATTERN = r'((?:\d+\s+\d+/\d+)|(?:\d+/\d+)|(?:\d+(?:[.,]\d+)?))'
UNIT_PATTERN = r'''('|"|d(?:[.]?|uymen)|pouc[e.]s?|in(?:ch(?:es)?|[.]?)|'''\
				r'''pieds?|v[.]?|voeten|f(?:eet|oot|t[.]?)|cm|lignes?|linges?)'''
DIMENSION_PATTERN = '(%s\\s*(?:%s)?)' % (NUMBER_PATTERN, UNIT_PATTERN)
DIMENSION_RE = re.compile(r'\s*%s' % (DIMENSION_PATTERN,))

SIMPLE_WIDTH_HEIGHT_PATTERN = r'(?:\s*((?<!\w)[wh]|width|height))?'
SIMPLE_DIMENSIONS_PATTERN_X1 = ''\
	r'(?P<d1>(?:%s\s*)+)(?P<d1w>%s)' % (DIMENSION_PATTERN, SIMPLE_WIDTH_HEIGHT_PATTERN)
SIMPLE_DIMENSIONS_RE_X1 = re.compile(SIMPLE_DIMENSIONS_PATTERN_X1)
SIMPLE_DIMENSIONS_PATTERN_X2a = ''\
	r'(?P<d1>(?:%s\s*)+)(?P<d1w>%s)'\
	r'(?:,)?\s*(x|by)'\
	r'(?P<d2>(?:\s*%s)+)(?P<d2w>%s)' % (
		DIMENSION_PATTERN,
		SIMPLE_WIDTH_HEIGHT_PATTERN,
		DIMENSION_PATTERN,
		SIMPLE_WIDTH_HEIGHT_PATTERN)
SIMPLE_DIMENSIONS_PATTERN_X2b = ''\
	r'(?P<d1w>%s)\s*(?P<d1>(?:%s\s*)+)'\
	r'*(?:(?:,)?\s*(x|by)|,\s*)'\
	r'(?P<d2w>%s)\s*(?P<d2>(?:\s*%s)+)' % (
		SIMPLE_WIDTH_HEIGHT_PATTERN,
		DIMENSION_PATTERN,
		SIMPLE_WIDTH_HEIGHT_PATTERN,
		DIMENSION_PATTERN)
SIMPLE_DIMENSIONS_RE_X2a = re.compile(SIMPLE_DIMENSIONS_PATTERN_X2a)
SIMPLE_DIMENSIONS_RE_X2b = re.compile(SIMPLE_DIMENSIONS_PATTERN_X2b)

# Haut 14 pouces, large 10 pouces
FRENCH_DIMENSIONS_PATTERN = r'[Hh](?:(?:aut(?:eur|[.])?)|[.])\s*(?P<d1>(?:%s\s*)+),? '\
							r'[Ll](?:(?:arge?(?:ur|[.])?)|[.])\s*(?P<d2>(?:%s\s*)+)' % (
								DIMENSION_PATTERN, DIMENSION_PATTERN)
FRENCH_DIMENSIONS_RE = re.compile(FRENCH_DIMENSIONS_PATTERN)

# Hoog. 1 v. 6 d., Breed 2 v. 3 d.
DUTCH_DIMENSIONS_PATTERN = r'(?P<d1w>[Hh]oogh?[.]?|[Bb]reedt?) (?P<d1>(?:%s\s*)+), '\
							r'(?P<d2w>[Hh]oogh?[.]?|[Bb]reedt?) (?P<d2>(?:%s\s*)+)' % (
								DIMENSION_PATTERN, DIMENSION_PATTERN)
DUTCH_DIMENSIONS_RE = re.compile(DUTCH_DIMENSIONS_PATTERN)

Dimension = namedtuple("Dimension", [
	'value',	# numeric value
	'unit',		# unit
	'which'		# e.g. width, height, ...
])

def _canonical_value(value):
	try:
		value = value.replace(',', '.')
		parts = value.split(None, 1)
		if len(parts) > 1 and '/' in parts[1]:
			intpart = int(parts[0])
			numer, denom = map(int, parts[1].split('/', 1))
			fracpart = float(numer)/denom
			value = str(intpart + fracpart)
		elif len(parts) == 1 and '/' in parts[0]:
			intpart = 0
			numer, denom = map(int, parts[0].split('/', 1))
			fracpart = float(numer)/denom
			value = str(intpart + fracpart)
		if value.startswith('.'):
			value = '0' + value
		
		try:
			return int(value)
		except ValueError:
			return float(value)
	except:
		pass
	return None

def _canonical_unit(value):
	inches = {'duymen', 'd.', 'd', '"'}
	feet = {'pieds', 'pied', 'feet', 'foot', 'voeten', 'v.', 'v', "'"}
	fr_inches = {'pouces', 'pouce', 'pouc.'}
	fr_feet = {'pieds', 'pied'}
	if value is None:
		return None
	value = value.lower()
	if value in fr_inches:
		return 'fr_inches'
	if value in fr_feet:
		return 'fr_feet'
	if 'ligne' in value or 'linge' in value:
		return 'ligne'
	if 'in' in value or value in inches:
		return 'inches'
	if 'ft' in value or value in feet:
		return 'feet'
	if 'cm' in value:
		return 'cm'
	return None

def _canonical_which(value):
	if not value:
		return None
	value = value.strip().lower()
	if value.startswith('w'):
		return 'width'
	if value.startswith('h'):
		return 'height'
	warnings.warn('*** Unknown which dimension: %s' % (value,))
	return None

def parse_simple_dimensions(value, which=None):
	'''
	Parse the supplied string for dimensions (value + unit), and return a list of
	`Dimension`s, optionally setting the `which` property to the supplied value.

	Examples:

	1 cm
	2ft
	5 pieds
	'''
	if value is None:
		return None
	value = value.strip()
	dims = []
# 	warnings.warn('DIMENSION: %s' % (value,))

	last_unit = None
	for match in re.finditer(DIMENSION_RE, value):
		# warnings.warn('--> match %s' % (match,))
		matched_value = _canonical_value(match.group(2))
		if not matched_value:
			warnings.warn('*** failed to canonicalize dimension value: %s' % (match.group(2),))
			return None
		unit_value = match.group(3)
		matched_unit = _canonical_unit(unit_value)
		if matched_unit is None:
			matched_unit = NEXT_FINER_DIMENSION_UNIT.get(last_unit)
		if unit_value and not matched_unit:
			warnings.warn('*** not a recognized unit: %s' % (unit_value,))
		which = _canonical_which(which)
		dim = Dimension(value=matched_value, unit=matched_unit, which=which)
		last_unit = matched_unit
		dims.append(dim)
	if not dims:
		return None
	return dims

def normalized_dimension_object(dimensions, source=None, **kwargs):
	'''
	Normalizes the given `dimensions`, or returns `None` is normalization fails.

	Returns a tuple of the normalized data, and a label which preserves the original
	set of dimensions.

	For example, the input:

		[
			Dimension(value=10, unit='feet', which=None),
			Dimension(value=3, unit='inches', which=None),
		]

	results in the output:

		(
			Dimension(value=123.0, unit='inches', which=None),
			"10 feet, 3 inches"
		)
	'''
	normalized = normalize_dimension(dimensions, source=source, **kwargs)
	if not normalized:
		return None
	labels = []
	for dim in dimensions:
		if dim.unit == 'inches':
			units = ('inch', 'inches')
		elif dim.unit == 'feet':
			units = ('foot', 'feet')
		elif dim.unit == 'fr_feet':
			units = ('French foot', 'French feet')
		elif dim.unit == 'fr_inches':
			units = ('French inch', 'French inches')
		elif dim.unit == 'cm':
			units = ('cm', 'cm')
		elif dim.unit == 'ligne':
			units = ('ligne', 'lignes')
		elif dim.unit is None:
			units = ('', '')
		else:
			warnings.warn('*** unrecognized unit: {dim.unit}')
			return None
		unit = units[0] if (float(dim.value) == 1.0) else units[1]
		if unit:
			label = '%s %s' % (dim.value, unit)
		else:
			label = str(dim.value)
		labels.append(label)
	label = ', '.join(labels)
	return normalized, label

def normalize_dimension(dimensions, source=None):
	'''
	Given a list of `Dimension`s, normalize them into a single Dimension (e.g. values in
	both feet and inches become a single dimension of inches).

	If the values cannot be sensibly combined (e.g. inches + centimeters), returns `None`.
	'''
	unknown = 0
	inches = 0
	fr_inches = 0
	centimeters = 0
	used_unknown = False
	used_inches = False
	used_fr_inches = False
	used_centimeters = False
	which = None
	for dim in dimensions:
		which = dim.which
		if dim.unit == 'inches':
			inches += dim.value
			used_inches = True
		elif dim.unit == 'feet':
			inches += 12 * dim.value
			used_inches = True
		elif dim.unit == 'cm':
			centimeters += dim.value
			used_centimeters = True
		elif dim.unit == 'fr_feet':
			fr_inches += 12.0 * dim.value
			used_fr_inches = True
		elif dim.unit == 'fr_inches':
			fr_inches += dim.value
			used_fr_inches = True
		elif dim.unit == 'ligne':
			fr_inches += dim.value / 12.0
			used_fr_inches = True
		elif dim.unit is None:
			unknown += dim.value
			used_unknown = True
		else:
			warnings.warn('*** unrecognized unit: %s' % (dim.unit,))
			return None

	used_systems = 0
	for used in (used_inches, used_fr_inches, used_centimeters, used_unknown):
		if used:
			used_systems += 1
	if used_systems != 1:
		if source:
			warnings.warn('*** dimension used a mix of metric, imperial, and/or unknown: '\
							'%r; source is %r' % (dimensions, source))
		else:
			warnings.warn('*** dimension used a mix of metric, imperial, and/or unknown: '\
							'%r' % (dimensions,))
		return None
	if fr_inches:
		return Dimension(value=fr_inches, unit='fr_inches', which=which)
	if inches:
		return Dimension(value=inches, unit='inches', which=which)
	if centimeters:
		return Dimension(value=centimeters, unit='cm', which=which)
	return Dimension(value=unknown, unit=None, which=which)

def extract_physical_dimensions(dimstr, **kwargs):
	dimensions = dimensions_cleaner(dimstr, **kwargs)
	if dimensions:
		for orig_d in dimensions:
			dimdata = normalized_dimension_object(orig_d, source=dimstr)
			if dimdata:
				dimension, label = dimdata
				if dimension.which == 'height':
					dim = vocab.Height(ident='')
				elif dimension.which == 'width':
					dim = vocab.Width(ident='')
				else:
					dim = vocab.PhysicalDimension(ident='')
				dim.value = dimension.value
				dim.identified_by = model.Name(ident='', content=label)
				unit = vocab.instances.get(dimension.unit)
				if unit:
					dim.unit = unit
				yield dim

def dimensions_cleaner(value, default_unit=None):
	'''
	Attempt to parse a set of dimensions from the given string.

	Returns a tuple of `pipeline.util.Dimension` objects if parsing succeeds,
	None otherwise.
	'''
	if value is None:
		return None
	cleaners = [
		simple_dimensions_cleaner_x2,
		french_dimensions_cleaner_x2,
		dutch_dimensions_cleaner_x2,
		simple_dimensions_cleaner_x1
	]
	for cleaner in cleaners:
		dimensions = cleaner(value)
		if dimensions:
			if default_unit:
				for i, d_list in enumerate(dimensions):
					for j, d in enumerate(d_list):
						if d.unit is None:
							dimensions[i][j] = Dimension(value=d.value, unit=default_unit, which=d.which)
			return dimensions
	return None

def french_dimensions_cleaner_x2(value):
	'''Attempt to parse 2 dimensions from a French-formatted string.'''
	# Haut 14 pouces, large 10 pouces

	match = FRENCH_DIMENSIONS_RE.match(value)
	if match:
		groups = match.groupdict()
		dim1 = parse_simple_dimensions(groups['d1'], 'h')
		dim2 = parse_simple_dimensions(groups['d2'], 'w')
		if dim1 and dim2:
			return (dim1, dim2)
		warnings.warn('dim1: %s %s h' % (dim1, groups['d1']))
		warnings.warn('dim2: %s %s w' % (dim2, groups['d2']))
		warnings.warn('*** Failed to parse dimensions: %s' % (value,))
	return None

def dutch_dimensions_cleaner_x2(value):
	'''Attempt to parse 2 dimensions from a Dutch-formatted string.'''
	# Hoog. 1 v. 6 d., Breed 2 v. 3 d.
	# Breedt 6 v., hoog 3 v

	match = DUTCH_DIMENSIONS_RE.match(value)
	if match:
		groups = match.groupdict()
		height = 'h'
		width = 'w'
		if 'breed' in groups['d1w'].lower():
			height, width = width, height

		dim1 = parse_simple_dimensions(groups['d1'], height)
		dim2 = parse_simple_dimensions(groups['d2'], width)
		if dim1 and dim2:
			return (dim1, dim2)
		warnings.warn('dim1: %s %s h' % (dim1, groups['d1']))
		warnings.warn('dim2: %s %s w' % (dim2, groups['d2']))
		warnings.warn('*** Failed to parse dimensions: %s' % (value,))
	return None

def simple_dimensions_cleaner_x1(value):
	'''Attempt to parse 1 dimension from a string.'''
	# 1 cm
	# 1' 2"
	# 1 ft. 2 in. h

	match = SIMPLE_DIMENSIONS_RE_X1.match(value)
	if match:
		groups = match.groupdict()
		dim1 = parse_simple_dimensions(groups['d1'], groups['d1w'])
		if dim1:
			return (dim1,)
	return None

def simple_dimensions_cleaner_x2(value):
	'''Attempt to parse 2 dimensions from a string.'''
	# 1 cm x 2 in
	# 1' 2" by 3 cm
	# 1 ft. 2 in. h by 3 cm w

	for pattern in (SIMPLE_DIMENSIONS_RE_X2a, SIMPLE_DIMENSIONS_RE_X2b):
		match = pattern.match(value)
		if match:
			groups = match.groupdict()
			dim1 = parse_simple_dimensions(groups['d1'], groups['d1w'])
			dim2 = parse_simple_dimensions(groups['d2'], groups['d2w'])
			if dim1 and dim2:
				return (dim1, dim2)
			warnings.warn('dim1: %s %s %s' % (dim1, groups['d1'], groups['d1w']))
			warnings.warn('dim2: %s %s %s' % (dim2, groups['d2'], groups['d2w']))
			warnings.warn('*** Failed to parse dimensions: %s' % (value,))
	return None

#mark - Monetary Values

def extract_monetary_amount(data, add_citations=False, currency_mapping=CURRENCY_MAPPING, source_mapping=None, truncate_label_digits=2):
	'''
	Returns a `MonetaryAmount`, `StartingPrice`, or `EstimatedPrice` object
	based on properties of the supplied `data` dict. If no amount or currency
	data is found in found, returns `None`.

	For `EstimatedPrice`, values will be accessed from these keys:
		- amount: `est_price_amount` or `est_price`
		- currency: `est_price_currency` or `est_price_curr`
		- note: `est_price_note` or `est_price_desc`
		- bibliographic statement: `est_price_citation`

	For `StartingPrice`, values will be accessed from these keys:
		- amount: `start_price_amount` or `start_price`
		- currency: `start_price_currency` or `start_price_curr`
		- note: `start_price_note` or `start_price_desc`
		- bibliographic statement: `start_price_citation`

	For `MonetaryAmount` prices, values will be accessed from these keys:
		- amount: `price_amount` or `price`
		- currency: `price_currency` or `price_curr`
		- note: `price_note` or `price_desc`
		- bibliographic statement: `price_citation`
	'''
	amount_type = 'Price'
	if 'price' in data or 'price_amount' in data or 'amount' in data:
		amnt = model.MonetaryAmount(ident='')
		price_amount = data.get('price_amount', data.get('price', data.get('amount')))
		price_currency = data.get('currency', data.get('price_currency', data.get('price_curr')))
		note = data.get('price_note', data.get('price_desc', data.get('note')))
		cite = data.get('price_citation', data.get('citation'))
		source = data.get('price_source', '')
	elif 'est_price' in data or 'est_price_amount' in data:
		amnt = vocab.EstimatedPrice(ident='')
		price_amount = data.get('est_price_amount', data.get('est_price'))
		price_currency = data.get('est_price_currency', data.get('est_price_curr', data.get('currency')))
		amount_type = 'Estimated Price'
		note = data.get('est_price_note', data.get('est_price_desc', data.get('note')))
		cite = data.get('est_price_citation', data.get('citation'))
		source = data.get('est_price_source', data.get('est_price_so', ''))
	elif 'start_price' in data or 'start_price_amount' in data:
		amnt = vocab.StartingPrice(ident='')
		price_amount = data.get('start_price_amount', data.get('start_price'))
		price_currency = data.get('start_price_currency', data.get('start_price_curr', data.get('currency')))
		amount_type = 'Starting Price'
		note = data.get('start_price_note', data.get('start_price_desc', data.get('note')))
		cite = data.get('start_price_citation', data.get('citation'))
		source = data.get('start_price_source', data.get('start_price_so', ''))
	elif 'ask_price' in data or 'ask_price_amount' in data:
		amnt = vocab.AskingPrice(ident='')
		price_amount = data.get('ask_price_amount', data.get('ask_price'))
		price_currency = data.get('ask_price_currency', data.get('ask_price_curr', data.get('currency')))
		amount_type = 'Asking Price'
		note = data.get('ask_price_note', data.get('ask_price_desc', data.get('note')))
		cite = data.get('ask_price_citation', data.get('citation'))
		source = data.get('ask_price_source', data.get('ask_price_so', ''))
	else:
		return None

	price_amount_label = price_amount
	if price_amount or price_currency:
		if cite and add_citations:
			amnt.referred_to_by = vocab.BibliographyStatement(ident='', content=cite)
		if note:
			amnt.referred_to_by = vocab.Note(ident='', content=note)

		if price_amount:
			try:
				value = price_amount
				value = value.replace('[?]', '')
				value = value.replace('?', '')
				value = value.strip()
				if re.search(re.compile(r',\d\d\d'), value):
					value = value.replace(',', '')
				value = float(value)
				
				label_fmt = '{:,.%df}' % truncate_label_digits
				price_amount_label = label_fmt.format(value)

				amnt.value = value
			except ValueError:
				amnt._label = price_amount
				amnt.identified_by = model.Name(ident='', content=price_amount)
	# 			warnings.warn(f'*** Not a numeric price amount: {value}')
		if price_currency:
			price_currency_key = price_currency
			try:
				price_currency_key = currency_mapping[price_currency_key.lower()]
			except KeyError:
				pass
			if isinstance(price_currency_key, model.BaseResource):
				amnt.currency = price_currency_key
			elif price_currency_key in vocab.instances:
				amnt.currency = vocab.instances[price_currency_key]
			else:
				warnings.warn('*** No currency instance defined for %s' % (price_currency_key,))
		if price_amount_label and price_currency:
			amnt._label = '%s %s' % (price_amount_label, price_currency)
		elif price_amount:
			amnt._label = '%s' % (price_amount,)
		return amnt
	return None



# Datetime Cleaning (from Getty Pipeline code)
# https://github.com/thegetty/pipeline/blob/master/pipeline/util/cleaners.py 

def date_parse(value, delim):
	# parse a / or - or . date or range

	bits = value.split(delim)
	if len(bits) == 2:
		# YYYY/ range
		b1 = bits[0].strip()
		b2 = bits[1].strip()
		if len(b2) < 3 :
			b2 = "%s%s" % (b1[:len(b1)-len(b2)], b2)
		elif len(b2) > 4:
			print("Bad range: %s" % value)
			return None
		try:
			return [datetime(int(b1),1,1), datetime(int(b2)+1,1,1)]
		except:
			print("Broken delim: %s" % value)
			return None
	elif len(bits) == 3:
		# YYYY/MM/DD or YY/YY/YYYY or DD.MM.YYYY or YYYY.MM.DD
		m = int(bits[1])
		if len(bits[0]) == 4:
			y = int(bits[0])
			d = int(bits[2])
		else:
			y = int(bits[2])
			d = int(bits[0])
		if m == 0:
			m = 1
		if d == 0:
			d = 1
		if m > 12:
			# swap them
			d, m = m, d
		try:
			yearmonthday = datetime(y,m,d)
			return [yearmonthday, yearmonthday+timedelta(days=1)]
		except:
			print("Bad // value: %s" % value)
	else:
		print("broken / date: %s" % value)
	return None

def date_cleaner(value):

	# FORMATS:

	# YYYY[?]
	# YYYY/MM/DD
	# DD/MM/YYYY
	# ca. YYYY
	# aft[er|.] YYYY
	# bef[ore|.] YYYY
	# YYYY.MM.DD
	# YYYY/(Y|YY|YYYY)
	# YYYY-YY
	# YYY0s
	# YYYY-
	# YYYY Mon
	# YYYY Month DD

	if value:
		value = value.replace("?",'')
		value = value.replace('est', '')
		value = value.replace("()", '')
		value = value.replace(' or ', '/')
		value = value.strip()
		value = value.replace('by ', 'bef.')
		value = value.replace('c.', 'ca.')
		value = value.replace('CA.', 'ca.')
		value = value.replace('af.', 'aft.')

	if not value:
		return None

	elif value.startswith("|"):
		# Broken? null it out
		return None

	elif len(value) <= 4 and value.isdigit():
		# year only
		return [datetime(int(value),1,1), datetime(int(value)+1,1,1)]

	elif value.startswith('v.'):
		value = value[2:].strip()
		return None

	elif value.startswith('-') and value[1:].isdigit():
		# BCE year
		# These are problematic, as python datetime.datetime() doesn't support them :(
		return [int(value), int(value)+1]

	elif value.endswith('s'):
		# 1950s
		if len(value) == 5 and value[:4].isdigit():
			y = int(value[:4])
			return [datetime(y,1,1), datetime(y+10,1,1)]
		else:
			warnings.warn("Bad YYYYs date: %s" % value)
			return None

	elif len(value) == 5 and value[:4].isdigit() and value.endswith('-'):
		y = int(value[:4])
		return [datetime(y,1,1), None]

	elif value.startswith("ca"):
		# circa x
		value = value[3:].strip()
		if len(value) == 4 and value.isdigit():
			y = int(value)
			return [datetime(y-CIRCA,1,1), datetime(y+CIRCA,1,1)]
		else:
			# Try and parse it
			if value.find('/') > -1:
				val = date_parse(value, '/')
			elif value.find('-') > -1:
				val = date_parse(value, '-')

			if not val:
				warnings.warn("bad circa: %s" % value)
				return None

			val[0] -= CIRCA_D
			val[1] += CIRCA_D
			return val

	elif value.startswith('aft'):
		# after x
		value = value.replace('aft.', '')
		value = value.replace('after ', '')
		value = value.strip()
		try:
			y = int(value)
		except:
			warnings.warn("Bad aft value: %s" % value)
			return None
		return [datetime(y,1,1), None]

	elif value.startswith('bef'):
		value = value.replace('bef.', '')
		value = value.replace('before ', '')
		value = value.strip()
		y = int(value)
		return [None, datetime(y,1,1)]

	elif value.find('/') > -1:
		# year/year or year/month/date
		# 1885/90
		# 07/02/1897
		return date_parse(value, '/')

	elif value.find('.') > -1:
		return date_parse(value, '.')

	elif value.find('-') > -1:
		# 0 could be -983 for 983 BCE
		return date_parse(value, '-')

	elif value.find(';') > -1:
		return date_parse(value, ';')

	else:
		with c_locale(), suppress(ValueError):
			yearmonthday = datetime.strptime(value, '%Y %B %d')
			if yearmonthday:
				return [yearmonthday, yearmonthday+timedelta(days=1)]

		with c_locale(), suppress(ValueError):
			yearmonth = datetime.strptime(value, '%Y %b')
			if yearmonth:
				year = yearmonth.year
				month = yearmonth.month
				maxday = calendar.monthrange(year, month)[1]
				d = datetime(year, month, 1)
				r = [d, d+timedelta(days=maxday)]
				return r

		warnings.warn('fell through to: {value!r}'.format(value=value))
		return None

@contextmanager
def c_locale():
	l = locale.getlocale()
	locale.setlocale(locale.LC_ALL, 'C')
	try:
		yield
	finally:
		locale.setlocale(locale.LC_ALL, l)




from cromulent import model, vocab
from cromulent.model import factory, DataError, OrderedDict, BaseResource
from cromulent.model import STR_TYPES

import json
import requests

class Reader(object):

	def __init__(self, validate_props=True, validate_profile=True):
		self.uri_object_map = {}
		self.forward_refs = []
		self.vocab_props = ['assigned_property']
		self.vocab_classes = {}
		self.validate_profile = validate_profile
		self.validate_props = validate_props

		for cx in dir(vocab):
			what = getattr(vocab, cx)
			# crying cat face -- type as a @property returns the function, not the value
			# when calling it on a class rather than an instance
			try:
				mytype = what._classhier[0].__name__
			except AttributeError:
				continue
			# find classes
			if (cx[0].isupper() and not hasattr(model, cx) and type(what) == type):
				# class
				self.vocab_classes[(mytype, what._classification[0].id)] = what

	def read(self, data):
		if not data:
			raise DataError("No data provided: %r" % data)
		elif type(data) in STR_TYPES:
			try:
				data = json.loads(data)
			except:
				raise DataError("Data is not valid JSON")
		if not data:
			raise DataError("No Data provided")
		self.uri_object_map = {}
		self.forward_refs = []
		try:
			what = self.construct(data)
			self.process_forward_refs()
			self.uri_object_map = {}
			self.forward_refs = []
			return what
		except:
			raise

	def process_forward_refs(self):
		for (what, prop, uri) in self.forward_refs:
			if uri in self.uri_object_map:
				setattr(what, prop, self.uri_object_map[uri])
			else:
				raise NotImplementedError("No class information for %s.%s = %s" % (what, prop, uri))

	def construct(self, js):
		# pass in json, get back object
		if '@context' in js:
			del js['@context']

		ident = js.get('id', '')
		typ = js.get('type', None)

		if typ == None:
			clx = BaseResource
		else:
			# Get class based on name
			try:
				clx = getattr(model, typ)
			except AttributeError:
				# No such class
				raise DataError("Resource %s has unknown class %s" % (ident, typ) )

		# now check vocab.ext_classes to try and refine
		trash = None 
		if 'classified_as' in js:
			for c in js['classified_as']:
				i = c.get('id', '')
				clx2 = self.vocab_classes.get((typ, i), None)
				if clx2 is not None:
					clx = clx2
					trash = c
					break

		what = clx(ident=ident)
		what._validate_profile = self.validate_profile
		self.uri_object_map[ident] = what

		if self.validate_props:
			propList = what.list_all_props()

		# sort data by KOH to minimize chance of bad backrefs
		itms = list(js.items())
		itms.sort(key=lambda x: factory.key_order_hash.get(x[0], 10000))

		for (prop, value) in itms:
			if prop in ['id', 'type']:
				continue

			if self.validate_props and not prop in propList:
				raise DataError("Unknown property %s on %s" % (prop, clx.__name__))

			# Climb looking for range
			rng = None
			for c in what._classhier:		
				if prop in c._all_properties:
					rng = c._all_properties[prop].range
					break

			if not rng:
				print(f"Couldn't find range for {prop} on {what}; trashing it")
				continue

			if type(value) != list:
				value = [value]
			for subvalue in value:
				if trash is not None and prop == 'classified_as' and subvalue == trash:
					continue
				if rng == str:
					setattr(what, prop, subvalue)				
				elif type(subvalue) == dict or isinstance(subvalue, OrderedDict):
					# recurse ...
					val = self.construct(subvalue)
					setattr(what, prop, val)
				elif type(subvalue) in STR_TYPES and prop in self.vocab_props:
					# keep as string
					setattr(what, prop, subvalue)
				elif type(subvalue) in STR_TYPES:
					# raw URI to be made into a class of type rng
					# or back reference
					if subvalue in self.uri_object_map:
						setattr(what, prop, self.uri_object_map[subvalue])
					elif rng in [model.Type, BaseResource]:
						# Always a X, often no more info
						setattr(what, prop, rng(ident=subvalue))
					else:
						self.forward_refs.append([what, prop, subvalue])
				else:
					# No idea!!
					raise DataError("Value %r is not expected for %s" % (subvalue, prop))

		return what


class NetworkedReader(Reader):
	# Read in data over the network, given a URI or JSON
	# Then traverse to n hops


	def __init__(self, validate_props=True, validate_profile=True):	
		Reader.__init__(self, validate_props, validate_profile)
		self.distances = {}
		self.results = {}
		self.ignore_props = ['access_point']


	def walk_for_refs(self, what, distance):
		props = what.list_my_props()
		if what.id and what.id.startswith('http') and not what.id in self.distances:
			self.distances[what.id] = distance
		for p in props:
			if not p in self.ignore_props:
				val = getattr(what, p)
				if isinstance(val, BaseResource):
					# walk
					self.walk_for_refs(val, distance)
				elif type(val) == list:
					for v in val:
						if isinstance(v, BaseResource):
							# walk
							self.walk_for_refs(v, distance)

	def read(self, data, hops=0):
		if type(data) in STR_TYPES and (data.startswith('http://') or \
				data.startswith('https://')):
			try:
				resp = requests.get(data)
				data = resp.json()				
			except:
				return DataError("Detected URI, but no data available")
		result = Reader.read(self, data)
		if hops:
			self.results[result.id] = result
			self.walk_for_refs(result, distance=hops-1)
			for uri in self.distances.keys():
				if not uri in self.results:
					self.read(uri, hops=hops-1)
		else:
			self.results[result.id] = result			
		return result



# This assumes the default CIDOC-CRM, even though the model code
# can generate classes for any ontology

import os
import json

from .model import (
    Identifier,
    Mark,
    HumanMadeObject,
    Type,
    Person,
    Material,
    MeasurementUnit,
    Place,
    Dimension,
    Currency,
    ConceptualObject,
    TimeSpan,
    Actor,
    PhysicalThing,
    Language,
    LinguisticObject,
    InformationObject,
    Formation,
    Dissolution,
    Activity,
    Group,
    Name,
    MonetaryAmount,
    Right,
    Encounter,
    Destruction,
    AttributeAssignment,
    BaseResource,
    PhysicalObject,
    Acquisition,
    HumanMadeFeature,
    VisualItem,
    Set,
    Birth,
    Death,
    PropositionalObject,
    Payment,
    Creation,
    Period,
    Production,
    Event,
    DigitalObject,
    TransferOfCustody,
    Move,
    DigitalService,
    CRMEntity,
    STR_TYPES,
    factory,
    ExternalResource,
)


# Add classified_as initialization hack for all resources
def post_init(self, **kw):
    if self.__class__._classification:
        for t in self._classification:
            self.classified_as = t


BaseResource._post_init = post_init

instances = {}
instance_types = {}


def register_vocab_class(name, data):
    parent = data["parent"]
    id = data["id"]
    label = data["label"]
    vocab = data.get("vocab", "aat")

    # We've imported them above...
    parent = globals()[parent]

    c = type(name, (parent,), {})
    if id.startswith("http"):
        t = Type(id)
    else:
        t = Type("http://vocab.getty.edu/%s/%s" % (vocab, id))
    t._label = label
    instance_types[name] = t
    c._classification = [t]
    if "metatype" in data:
        t.classified_as = instances[data["metatype"]]
    c._type = None  # To avoid conflicting with parent class
    globals()[name] = c
    return c


def register_instance(name, data):
    parent = data["parent"]
    id = data["id"]
    vocab = data.get("vocab", "aat")
    label = data["label"]

    parent = globals()[parent]

    if id.startswith("http"):
        t = parent(id)
    else:
        t = parent("http://vocab.getty.edu/%s/%s" % (vocab, id))
    t._label = label
    instances[name] = t

    # Languages have a code (e.g. 'en') that goes in `notation`
    if 'code' in data:
        t.notation = data['code']

    return t


# Consider:
# pen, pencil, card, cardboard, porcelain, wax, ceramic, plaster
# crayon, millboard, gouache, brass, stone, lead, iron, clay,
# alabaster, limestone

# Read in external files from data/ directory
dd = os.path.join(os.path.dirname(__file__), "data")
cls_fn = os.path.join(dd, "vocab_classes.json")
inst_fn = os.path.join(dd, "vocab_instances.json")

fh = open(cls_fn)
cls_data = fh.read()
fh.close()
fh = open(inst_fn)
inst_data = fh.read()
fh.close()

cls_js = json.loads(cls_data)
inst_js = json.loads(inst_data)

for name, v in inst_js.items():
    if "parent" in v and not v["parent"] in cls_js:
        register_instance(name, v)
for name, v in cls_js.items():
    if "parent" in v:
        register_vocab_class(name, v)
for name, v in inst_js.items():
    if "parent" in v and v["parent"] in cls_js:
        register_instance(name, v)


def add_classification(obj, cl_type):
    c = cl_type()
    for cn in c._classification:
        if hasattr(obj, "classified_as") and not cn in obj.classified_as:
            obj.classified_as = cn
        elif not hasattr(obj, "classified_as"):
            obj.classified_as = cn
    return obj


def make_multitype_obj(*args, **kw):
    # (class1, class2, class3, name=foo, other=bar)
    inst = args[0](**kw)
    for c in args[1:]:
        for cn in c._classification:
            if hasattr(inst, "classified_as") and not cn in inst.classified_as:
                inst.classified_as = cn
            elif not hasattr(inst, "classified_as"):
                inst.classified_as = cn
    return inst


def conceptual_only_parts():
    # Make .part work as expected for Right
    # which is only propositional and not symbolic, so P148 not P106

    def set_conceptual_part(self, value):
        self.conceptual_part = value

    def set_conceptually_part_of(self, value):
        self.conceptually_part_of = value

    def rights_getter(self, what):
        if what == "part":
            return self.conceptual_part
        elif what == "part_of":
            return self.conceptually_part_of
        else:
            object.__getattr__(self, what)

    Right.set_part = set_conceptual_part
    Right.set_part_of = set_conceptually_part_of
    Right._property_name_map["conceptual_part"] = "part"
    Right._property_name_map["conceptually_part_of"] = "part_of"
    Right._all_properties["part"] = PropositionalObject._all_properties["conceptual_part"]
    Right._all_properties["part_of"] = PropositionalObject._all_properties["conceptually_part_of"]
    Right.__getattr__ = rights_getter


def add_art_setter():
    # Linked.Art profile requires aat:300133025 on all artworks
    # Art can be a HumanMadeObject or an InformationObject
    # set it by adding art=1 to the constructor

    def art_post_init(self, **kw):
        super(HumanMadeObject, self)._post_init(**kw)
        if "art" in kw:
            self.classified_as = instances["artwork"]

    HumanMadeObject._post_init = art_post_init

    def art2_post_init(self, **kw):
        if "art" in kw:
            self.classified_as = instances["artwork"]
        super(InformationObject, self)._post_init(**kw)

    InformationObject._post_init = art2_post_init


def add_attribute_assignment_check():
    # Allow references to properties in p177 on AttrAssign
    # Validate that the property is allowed in assigned
    # either on set, or when assigned is set

    p177 = factory.context_rev.get("crm:P177_assigned_property", "assigned_property")
    ass = factory.context_rev.get("crm:P141_assigned", "assigned")
    assto = factory.context_rev.get("crm:P140:assigned_attribute_to", "assigned_to")

    def aa_set_assigned(self, value):
        assto_res = getattr(self, assto, None)
        if assto_res:
            p177_res = getattr(self, p177, None)
            assto_res._check_prop(p177_res, value)

        current = getattr(self, ass, None)
        if current:
            value = [*current, value]
        elif type(value) is not list:
            value = [value]
        object.__setattr__(self, ass, value)

    setattr(AttributeAssignment, "set_%s" % ass, aa_set_assigned)

    def aa_set_assigned_to(self, value):
        ass_res = getattr(self, ass, None)
        p177_res = getattr(self, p177, None)
        if ass_res and p177_res:
            # unmap the URI to property name
            for ar in ass_res:
                value._check_prop(p177_res, ar)
        object.__setattr__(self, assto, value)

    setattr(AttributeAssignment, "set_%s" % assto, aa_set_assigned_to)

    def aa_set_assigned_property_type(self, value):
        ass_res = getattr(self, ass, None)
        assto_res = getattr(self, assto, None)
        if ass_res and assto_res:
            for ar in ass_res:
                assto_res._check_prop(value, ar)
        object.__setattr__(self, p177, value)

    setattr(AttributeAssignment, "set_%s" % p177, aa_set_assigned_property_type)


def add_linked_art_boundary_check():
    boundary_classes = [
        x.__name__
        for x in [
            Actor,
            HumanMadeObject,
            Person,
            Group,
            VisualItem,
            Place,
            Period,
            LinguisticObject,
            PropositionalObject,
            Set,
            Event,
            DigitalObject,
            DigitalService,
        ]
    ]
    data_embed_classes = [Name, Identifier, Dimension, TimeSpan, MonetaryAmount]
    type_embed_classes = [Type, Currency, Language, Material, MeasurementUnit]
    event_embed_classes = [Birth, Creation, Production, Formation, Payment, Death, Destruction, Dissolution]
    all_embed_classes = []
    all_embed_classes.extend(data_embed_classes)
    all_embed_classes.extend(type_embed_classes)
    all_embed_classes.extend(event_embed_classes)
    embed_classes = [x.__name__ for x in all_embed_classes]

    # Activity, AttributeAssignment, InformationObject, TransferOfCustody, Move
    # Propositional Object

    ExternalResource._embed_override = None

    def my_linked_art_boundary_check(self, top, rel, value):
        # True = Embed ; False = Split
        if value._embed_override is not None:
            return value._embed_override
        elif isinstance(value, LinguisticObject) and hasattr(value, "classified_as"):
            for ca in value.classified_as:
                if instances["brief text"] in getattr(ca, "classified_as", []):
                    return True
        # Non Statement Linguistic objects might still be internal or external
        # so apply logic from relating properties, not return False
        elif isinstance(value, ProvenanceEntry):
            return False

        boundary_crossing_props = set(
            [
                "part_of",
                "member_of",
                "specific_purpose",
                "caused_by",
                "starts_before_the_end_of",
                "ends_after_the_start_of",
                "starts_before_the_start_of",
                "starts_after_the_start_of",
                "ends_before_the_start_of",
                "starts_after_the_end_of",
                "ends_before_the_end_of",
                "ends_after_the_end_of",
            ]
        )

        if rel in ["part", "member", "specific_purpose_of", "caused"]:
            # Downwards, internal simple partitioning
            # This catches an internal part to a LinguisticObject
            return True
        elif rel in boundary_crossing_props:
            # upwards partition refs are inclusion, and always boundary crossing
            return False
        elif value.type in boundary_classes:
            # This catches the external text LinguisticObject
            return False
        elif value.type in embed_classes:
            return True
        else:
            # Default to embedding to avoid data loss
            return True

    setattr(BaseResource, "_linked_art_boundary_okay", my_linked_art_boundary_check)
    factory.linked_art_boundaries = True


def set_linked_art_uri_segments():
    HumanMadeObject._uri_segment = "object"
    Activity._uri_segment = "event"
    Event._uri_segment = "event"
    Period._uri_segment = "event"
    Place._uri_segment = "place"
    InformationObject._uri_segment = "info"
    Group._uri_segment = "group"
    # Actor._uri_segment = "actor"
    Person._uri_segment = "person"
    PhysicalObject._uri_segment = "object"
    LinguisticObject._uri_segment = "text"
    PropositionalObject._uri_segment = "concept"
    DigitalObject._uri_segment = "digital"
    DigitalService._uri_segment = "digital"
    Type._uri_segment = "concept"
    Language._uri_segment = "concept"
    MeasurementUnit._uri_segment = "concept"
    Currency._uri_segment = "concept"
    Material._uri_segment = "concept"
    VisualItem._uri_segment = "visual"
    ProvenanceEntry._uri_segment = "provenance"
    Exhibition._uri_segment = "activity"
    Set._uri_segment = "set"


def add_helper_functions():
    # Add filter functions to the right bits of the model

    def get_names(self, filter=None):
        return [x for x in self.identified_by if isinstance(x, Name) and (not filter or filter in x.classified_as)]

    def get_identifiers(self, filter=None):
        return [x for x in self.identified_by if isinstance(x, Identifier) and (not filter or filter in x.classified_as)]

    def get_statements(self, filter=None):
        return [
            x
            for x in self.referred_to_by
            if isinstance(x, LinguisticObject) and x.content and (not filter or filter in x.classified_as)
        ]

    CRMEntity.get_names = get_names
    CRMEntity.get_identifiers = get_identifiers
    CRMEntity.get_statements = get_statements
